"""

Steps:
 - make an image with borealis:
        `python borealis/gce.py -o image=vivarium-profiling-image3
        -o custom-cpu=8 -o custom-memory=8GB -l '' -m 'metadata=True'
        vivarium-profile-8cpus`
 - ssh into new image: `gcloud compute ssh vivarium-profile-8cpus-0`
 - run profile_image.py in the VM instance:
        `python3 vivarium/experiments/profile_image.py`
 - exit out of VM (ctrl + D) and retrieve results:
        `gcloud compute scp vivarium-profile-8cpus-0:
        vivarium-core/out/vm/scan_results.json
        out/vm/8cpu_results.json`

"""
import argparse
import json
import os
import multiprocessing

from vivarium.core.composition import BASE_OUT_DIR
from vivarium.experiments.profile_runtime import (
    ComplexModelSim, run_scan, plot_scan_results)

VM_OUT_DIR = os.path.join(BASE_OUT_DIR, 'vm')


def scan_parallel_save() -> None:
    n_cpus = multiprocessing.cpu_count()

    total_processes = 50
    scan_values = [
        {
            'number_of_processes': total_processes,
            'number_of_parallel_processes': total_processes,
            'number_of_cpus': n_cpus,
        }
    ]

    sim = ComplexModelSim()
    sim.process_sleep = 1e-2
    sim.experiment_time = 50
    saved_stats = run_scan(sim,
                           scan_values=scan_values)

    # save results
    os.makedirs(VM_OUT_DIR, exist_ok=True)
    fig_path = os.path.join(VM_OUT_DIR, 'scan_results.json')
    with open(fig_path, 'w') as outfile:
        json.dump(saved_stats, outfile)


def plot_vm_scan_results() -> None:
    path_to_json = VM_OUT_DIR

    # finds json files
    json_files = [
        pos_json for pos_json in os.listdir(path_to_json)
        if pos_json.endswith('.json')]

    # load json
    saved_stats = []
    for js in json_files:
        with open(os.path.join(path_to_json, js)) as json_file:
            json_text = json.load(json_file)
            saved_stats.extend(json_text)

    number_of_processes = None
    for saved_stat in saved_stats:
        if number_of_processes:
            assert number_of_processes == saved_stat['number_of_processes'], \
                "all scans must have the same number of processes"
        else:
            number_of_processes = saved_stat['number_of_processes']

    # plot
    plot_scan_results(saved_stats,
                      cpus_plot=True,
                      row_height=2.5,
                      out_dir=path_to_json,
                      title=f'n vCPUs running {number_of_processes} '
                            f'parallel processes',
                      filename='scan_vCPUs')


# python vivarium/experiments/profile_image.py
if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='profile vm instances')
    parser.add_argument(
        '--plot', '-p', action='store_true')
    parser_args = parser.parse_args()

    if parser_args.plot:
        plot_vm_scan_results()
    else:
        scan_parallel_save()



"""
==================
GrowthRate Process
==================
"""

import os
import numpy as np

from vivarium.core.process import Process
from vivarium.core.composition import (
    PROCESS_OUT_DIR,
    process_in_experiment,
    simulate_experiment,
)
from vivarium.plots.simulation_output import plot_simulation_output

NAME = 'growth_rate'


class GrowthRate(Process):
    """ A Vivarium process that models exponential growth of biomass """

    name = NAME
    defaults = {
        'default_growth_rate': 0.0005,
        'default_growth_noise': 0.0,
        'variables': ['mass']
    }

    def ports_schema(self):
        return {
            'variables': {
                variable: {
                    '_default': 1.0,
                    '_divider': 'split',
                    '_emit': True,
                } for variable in self.parameters['variables']
            },
            'rates': {
                'growth_rate': {
                    variable: {
                        '_default': self.parameters[
                            'default_growth_rate'],
                    } for variable in self.parameters['variables']
                },
                'growth_noise': {
                    variable: {
                        '_default': self.parameters[
                            'default_growth_noise'],
                    } for variable in self.parameters['variables']
                },
            }
        }

    def next_update(self, timestep, states):
        variables = states['variables']
        growth_rate = states['rates']['growth_rate']
        growth_noise = states['rates']['growth_noise']

        variable_update = {
            variable: value * np.exp(
                (
                    growth_rate[variable]
                    + np.random.normal(0, growth_noise[variable])
                ) * timestep
            ) - value
            for variable, value in variables.items()
        }
        return {'variables': variable_update}


def test_growth_rate(total_time=1350):
    initial_mass = 100
    growth_rate = 0.0005
    config = {
        'variables': ['mass'],
        'default_growth_rate': growth_rate,
        'time_step': 2,
    }

    growth_rate_process = GrowthRate(config)
    initial_state = {'variables': {'mass': initial_mass}}
    experiment = process_in_experiment(
        growth_rate_process,
        initial_state=initial_state)
    output = simulate_experiment(experiment, {'total_time': total_time})

    # asserts
    final_mass = output['variables']['mass'][-1]
    expected_mass = initial_mass * np.exp(growth_rate * total_time)
    decimal_precision = 11
    assert abs(expected_mass - final_mass) < \
           1.5 * 10 ** (-decimal_precision)
    return output


def main():
    """run test and plot"""
    out_dir = os.path.join(PROCESS_OUT_DIR, NAME)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    data = test_growth_rate()
    plot_settings = {}
    plot_simulation_output(data, plot_settings, out_dir)


if __name__ == '__main__':
    main()



import math
from typing import Optional, Dict, Any, Union

import numpy as np

from vivarium.core.process import (
    Process, Deriver, Step)
from vivarium.core.composer import Composer, MetaComposer, Composite
from vivarium.core.types import State, Schema, Update, Topology
from vivarium.processes.division import get_divide_update

quark_colors = ['green', 'red', 'blue']
quark_spins = ['up', 'down']
electron_spins = ['-1/2', '1/2']
electron_orbitals = [
    str(orbit) + 's'
    for orbit in range(1, 8)]


class ToyTransport(Process):
    name = 'toy_transport'
    defaults = {'intake_rate': 2}

    def __init__(
            self,
            parameters: Optional[Dict[str, Any]] = None
    ):
        super().__init__(parameters)

    def ports_schema(self):
        return {
            'external': {
                'GLC': {
                    '_default': 0.0,
                    '_emit': True,
                    '_divider': 'set'}},
            'internal': {
                'GLC': {
                    '_default': 0.0,
                    '_emit': True,
                    '_divider': 'split'}}}

    def next_update(self, timestep, states):
        update = {}
        intake = timestep * self.parameters['intake_rate']
        if states['external']['GLC'] >= intake:
            update = {
                'external': {'GLC': -2, 'MASS': 1},
                'internal': {'GLC': 2}}

        return update


class ToyDeriveVolume(Deriver):
    name = 'toy_derive_volume'

    def __init__(
            self,
            initial_parameters: Optional[Dict[str, Any]] = None
    ):
        _ = initial_parameters  # ignore initial_parameters
        parameters: Dict[str, Any] = {}
        super().__init__(parameters)

    def ports_schema(self):
        ports = {
            'compartment': ['MASS', 'DENSITY', 'VOLUME']}
        return {
            port_id: {
                key: {
                    '_updater': 'set' if key == 'VOLUME' else
                    'accumulate',
                    '_default': 0.0,
                    '_emit': True}
                for key in keys}
            for port_id, keys in ports.items()}

    def next_update(self, timestep, states):
        volume = states['compartment']['MASS'] /\
                 states['compartment']['DENSITY']
        update = {
            'compartment': {'VOLUME': volume}}

        return update


class ToyDeath(Process):
    name = 'toy_death'

    def __init__(
            self,
            initial_parameters: Optional[Dict[str, Any]] = None
    ):
        initial_parameters = initial_parameters or {}
        self.targets = initial_parameters.get('targets', [])
        super().__init__({})

    def ports_schema(self):
        return {
            'compartment': {
                'VOLUME': {
                    '_default': 0.0,
                    '_emit': True}},
            'global': {
                target: {
                    '_default': None}
                for target in self.targets}}

    def next_update(self, timestep, states):
        volume = states['compartment']['VOLUME']
        update = {}

        if volume > 1.0:
            # kill the cell
            update = {
                'global': {
                    '_delete': [
                        (target,)
                        for target in self.targets]}}

        return update


class ToyEnvironment(Process):
    port_ids = ['external', 'membrane']

    def ports_schema(self):
        return {
            'agents': {
                '*': {
                    port_id: {
                        '_default': 0.0
                    } for port_id in self.port_ids
                }
            }
        }

    def next_update(self, timestep, states):
        agents = states['agents']

        agents_update = {}
        for agent_id, agent_state in agents.items():
            assert set(agent_state.keys()) == set(self.port_ids), \
                'view is getting states not in ports_schema'
            agents_update[agent_id] = {}
            agents_update[agent_id]['external'] = 1

        return {'agents': agents_update}


class ToyCompartment(Composer):
    '''
    a toy compartment for testing

    '''
    def __init__(self, config):
        super().__init__(config)

    def generate_processes(self, config):
        return {
            'metabolism': ToyMetabolism(
                {'mass_conversion_rate': 0.5}),
            'transport': ToyTransport(),
            'death': ToyDeath({'targets': [
                'metabolism',
                'transport']}),
            'external_volume': ToyDeriveVolume(),
            'internal_volume': ToyDeriveVolume()
        }

    def generate_topology(self, config):
        return{
            'metabolism': {
                'pool': ('cytoplasm',)},
            'transport': {
                'external': ('periplasm',),
                'internal': ('cytoplasm',)},
            'death': {
                'global': tuple(),
                'compartment': ('cytoplasm',)},
            'external_volume': {
                'compartment': ('periplasm',)},
            'internal_volume': {
                'compartment': ('cytoplasm',)}}


class ToyMetabolism(Process):
    name = 'toy_metabolism'

    def __init__(
            self,
            initial_parameters: Optional[Dict[str, Any]] = None
    ):
        initial_parameters = initial_parameters or {}
        parameters = {'mass_conversion_rate': 1}
        parameters.update(initial_parameters)
        super().__init__(parameters)

    def ports_schema(self):
        ports = {
            'pool': ['GLC', 'MASS']}
        return {
            port_id: {
                key: {
                    '_default': 0.0,
                    '_emit': True}
                for key in keys}
            for port_id, keys in ports.items()}

    def next_update(self, timestep, states):
        update = {}
        glucose_required = (timestep /
                            self.parameters['mass_conversion_rate'])
        if states['pool']['GLC'] >= glucose_required:
            update = {
                'pool': {
                    'GLC': -2,
                    'MASS': 1}}

        return update


class ToyLinearGrowthDeathProcess(Process):

    name = 'toy_linear_growth_death'

    GROWTH_RATE = 1.0
    THRESHOLD = 6.0

    def __init__(
            self,
            initial_parameters: Optional[Dict[str, Any]] = None
    ):
        initial_parameters = initial_parameters or {}
        self.targets = initial_parameters.get('targets')
        super().__init__(initial_parameters)

    def ports_schema(self):
        return {
            'global': {
                'mass': {
                    '_default': 1.0,
                    '_emit': True}},
            'targets': {
                target: {
                    '_default': None}
                for target in self.targets}}

    def next_update(self, timestep, states):
        mass = states['global']['mass']
        mass_grown = (
            ToyLinearGrowthDeathProcess.GROWTH_RATE * timestep)
        update = {
            'global': {'mass': mass_grown},
        }
        if mass > ToyLinearGrowthDeathProcess.THRESHOLD:
            update['global'] = {
                '_delete': [(target,) for target in self.targets]}

        return update


class Proton(Process):
    name = 'proton'
    defaults = {
        'time_step': 1.0,
        'radius': 0.0}

    def ports_schema(self):
        return {
            'radius': {
                '_updater': 'set',
                '_default': self.parameters['radius']},
            'quarks': {
                '_divider': 'split_dict',
                '*': {
                    'color': {
                        '_updater': 'set',
                        '_default': quark_colors[0]},
                    'spin': {
                        '_updater': 'set',
                        '_default': quark_spins[0]}}},
            'electrons': {
                '*': {
                    'orbital': {
                        '_updater': 'set',
                        '_default': electron_orbitals[0]},
                    'spin': {
                        '_default': electron_spins[0]}}}}

    def next_update(self, timestep, states):
        update = {}

        collapse = np.random.uniform()
        if collapse < states['radius'] * timestep:
            update['radius'] = collapse
            update['quarks'] = {}

            for name in states['quarks'].keys():
                update['quarks'][name] = {
                    'color': np.random.choice(quark_colors),
                    'spin': np.random.choice(quark_spins)}

            update['electrons'] = {}
            orbitals = electron_orbitals.copy()
            for name in states['electrons'].keys():
                np.random.shuffle(orbitals)
                update['electrons'][name] = {
                    'orbital': orbitals.pop()}

        return update


class Electron(Process):
    name = 'electron'
    defaults = {
        'time_step': 1.0,
        'spin': electron_spins[0]}

    def ports_schema(self):
        return {
            'spin': {
                '_updater': 'set',
                '_default': self.parameters['spin']},
            'proton': {
                'radius': {
                    '_default': 0.0}}}

    def next_update(self, timestep, states):
        update = {}

        if np.random.uniform() < states['proton']['radius']:
            update['spin'] = np.random.choice(electron_spins)

        return update


class Sine(Process):
    name = 'sine'
    defaults = {
        'initial_phase': 0.0}

    def ports_schema(self):
        return {
            'frequency': {
                '_default': 440.0},
            'amplitude': {
                '_default': 1.0},
            'phase': {
                '_default': self.parameters['initial_phase']},
            'signal': {
                '_default': 0.0,
                '_updater': 'set'}}

    def next_update(self, timestep, states):
        phase_shift = timestep * states['frequency'] % 1.0
        signal = states['amplitude'] * math.sin(
            2 * math.pi * (states['phase'] + phase_shift))

        return {
            'phase': phase_shift,
            'signal': signal}


class ExchangeA(Process):
    """ Exchange A

    A minimal exchange process that moves molecules 'A' between internal
    and external ports
    """
    name = 'exchange_a'
    defaults = {
        'uptake_rate': 0.0,
        'secrete_rate': 0.0}

    def __init__(self, parameters=None):
        super().__init__(parameters)
        self.uptake_rate = self.parameters['uptake_rate']
        self.secrete_rate = self.parameters['secrete_rate']

    def ports_schema(self):
        return {
            'internal': {
                'A': {
                    '_default': 0.0,
                    '_emit': True}},
            'external': {
                'A': {
                    '_default': 0.0,
                    '_emit': True}}}

    def next_update(self, timestep, states):
        a_in = states['internal']['A']
        a_out = states['external']['A']
        delta_a_in = a_out * self.uptake_rate - a_in * self.secrete_rate
        return {
            'internal': {'A': delta_a_in},
            'external': {'A': -delta_a_in}}


class Po(Process):
    name = 'po'

    def initial_state(self, config=None):
        return {
            'A': {
                'a1': -1,
                'a2': -2,
                'a3': -3},
            'B': {
                'b1': -4,
                'b2': -5}}

    def ports_schema(self):
        return {
            'A': {
                'a1': {'_default': 0},
                'a2': {'_default': 0},
                'a3': {'_default': 0}},
            'B': {
                'b1': {'_default': 0},
                'b2': {'_default': 0}}}

    def next_update(self, timestep, states):
        return {
            'A': {
                'a1': 1,
                'a2': 1,
                'a3': 1},
            'B': {
                'b1': -1,
                'b2': -1}}


class Qo(Process):
    name = 'qo'

    def initial_state(self, config=None):
        return {
            'D': {
                'd1': 1,
                'd2': 2,
                'd3': 3},
            'E': {
                'e1': 4,
                'e2': 5}}

    def ports_schema(self):
        return {
            'D': {
                'd1': {'_default': 0},
                'd2': {'_default': 0},
                'd3': {'_default': 0}},
            'E': {
                'e1': {'_default': 0},
                'e2': {'_default': 0}}}

    def next_update(self, timestep, states):
        return {
            'D': {
                'd1': 10,
                'd2': 10,
                'd3': 10},
            'E': {
                'e1': -10,
                'e2': -10}}


class PoQo(Composer):
    def generate_processes(self, config=None):
        return {
            'po': Po(config),
            'qo': Qo(config),
        }

    def generate_topology(self, config=None):
        return {
            'po': {
                'A': {
                    '_path': ('aaa',),
                    'a2': ('x',),
                    'a3': ('..', 'ccc', 'a3')},
                'B': ('bbb',),
            },
            'qo': {
                'D': {
                    '_path': (),
                    'd1': ('aaa', 'd1'),
                    'd2': ('aaa', 'd2'),
                    'd3': ('ccc', 'd3')},
                'E': {
                    '_path': (),
                    'e1': ('aaa', 'x'),
                    'e2': ('bbb', 'e2')}
            },
        }


def test_composite_initial_state_complex() -> None:

    outer_path = ('universe', 'agent')
    pq = PoQo({})
    pq_composite = pq.generate(path=outer_path)
    pq_initial = pq_composite.initial_state(
        config={
            'initial_state': {
                'universe': {'agent': {'aaa': {'x': 4}}}}})

    expected_initial = {
        'universe': {
            'agent': {
                'aaa': {'a1': -1, 'd1': 1, 'd2': 2, 'x': 4},
                'bbb': {'b1': -4, 'b2': -5, 'e2': 5},
                'ccc': {'a3': -3, 'd3': 3}
            }}}

    assert pq_initial == expected_initial


class AA(Process):
    name = 'AA'

    def initial_state(self, config: Optional[dict] = None) -> State:
        return {'a_port': {'a': 1}}

    def ports_schema(self) -> Schema:
        return {'a_port': {'a': {'_emit': True}}}

    def next_update(
            self,
            timestep: Union[float, int],
            states: State) -> Update:
        return {'a_port': {'a': 1}}


class BB(Composer):
    name = 'BB'

    def generate_processes(
            self, config: Optional[dict]) -> Dict[str, Any]:
        return {
            'a1': AA({}),
            'a2': AA({}),
            'a3_store': {
                'a3': AA({})}}

    def generate_topology(self, config: Optional[dict]) -> Topology:
        return {
            'a1': {
                'a_port': ('a1_store',)},
            'a2': {
                'a_port': {
                    'a': ('a1_store', 'b')}},
            'a3_store': {
                'a3': {
                    'a_port': ('a3_1_store',)}}}


class ToyProcess(Process):
    name = 'toy'

    def ports_schema(self) -> Schema:
        return {
            'port1': {
                'var_a': {'_default': 0, '_emit': True},
                'var_b': {'_default': 0, '_emit': True}},
            'port2': {
                'var_a': {'_default': 0, '_emit': True},
                'var_b': {'_default': 0, '_emit': True}}}

    def next_update(
            self, timestep: Union[float, int], states: State) -> Update:
        return {
            'port1': {
                'var_a': 1,
                'var_b': states['port1']['var_a']},
            'port2': {
                'var_a': states['port1']['var_b'],
                'var_b': states['port2']['var_a']}}


class ToyComposer(Composer):
    defaults = {
        'process1':  {'name': 'process1'},
        'process2': {'name': 'process2'}}

    def generate_processes(
            self,
            config: Optional[dict]
    ) -> Dict[str, ToyProcess]:
        assert config is not None
        config = config or self.defaults
        process1 = ToyProcess(config['process1'])
        process2 = ToyProcess(config['process2'])
        return {
            process1.name: process1,
            process2.name: process2}

    def generate_topology(
            self,
            config: Optional[dict] = None
    ) -> Topology:
        config = config or self.defaults
        a_name = config['process1']['name']
        b_name = config['process2']['name']
        return {
            a_name: {
                'port1': ('store_A',),
                'port2': ('store_B',)},
            b_name: {
                'port1': ('store_B',),
                'port2': ('store_C',)}}


def test_override() -> None:
    config = {
        '_schema': {
            'a3_store': {
                'a3': {
                    'a_port': {
                        'a': {
                            '_default': 2}}}}}}
    bb_composer = BB(config)
    bb_composite = bb_composer.generate()
    default_state = bb_composite.default_state()

    expected_default_state = {
        'a3_store': {
            'a3_1_store': {
                'a': 2}}}

    assert default_state == expected_default_state


def test_composite_initial_state() -> None:
    """
    test that initial state in composite merges individual processes'
    initial states
    """

    bb_composer = BB({})
    bb_composite = bb_composer.generate()

    composer_initial_state = bb_composer.initial_state()
    composite_initial_state = bb_composite.initial_state()

    expected_initial_state = {
        'a3_store': {
            'a3_1_store': {
                'a': 1}},
        'a1_store': {
            'a': 1,
            'b': 1}}

    assert composite_initial_state == composer_initial_state
    assert composite_initial_state == expected_initial_state


def test_composite_parameters() -> None:
    """
    test that initial state in composite merges individual processes'
    initial states
    """

    bb_composer = BB({})
    bb_composite = bb_composer.generate()
    composer_parameters = bb_composer.get_parameters()
    composite_parameters = bb_composite.get_parameters()
    expected_parameters = {
        'a1': {'time_step': 1.0},
        'a2': {'time_step': 1.0},
        'a3_store': {
            'a3': {'time_step': 1.0}}}
    assert composite_parameters == composer_parameters
    assert composite_parameters == expected_parameters


def test_composite_merge() -> None:
    composer = ToyComposer()
    composite = composer.generate()

    expected_initial_topology = {
        'process1': {
            'port1': ('store_A',),
            'port2': ('store_B',),
        },
        'process2': {
            'port1': ('store_B',),
            'port2': ('store_C',),
        },
    }
    assert composite['topology'] == expected_initial_topology

    for key in ('process1', 'process2'):
        assert key in composite['processes']
        assert isinstance(composite['processes'][key], ToyProcess)

    # merge
    merge_processes: Dict[str, Process] = {
        'process3': ToyProcess({'name': 'process3'})}
    merge_topology: Topology = {
        'process3': {
            'port1': ('store_A',),
            'port2': ('store_B',)}}
    composite.merge(
        processes=merge_processes,
        topology=merge_topology)

    expected_merged_topology = {
        'process1': {
            'port1': ('store_A',),
            'port2': ('store_B',),
        },
        'process2': {
            'port1': ('store_B',),
            'port2': ('store_C',),
        },
        'process3': {
            'port1': ('store_A',),
            'port2': ('store_B',),
        },
    }
    assert composite['topology'] == expected_merged_topology

    for key in ('process1', 'process2', 'process3'):
        assert key in composite['processes']
        assert isinstance(composite['processes'][key], ToyProcess)


def test_get_composite() -> None:
    process1 = ToyProcess({'name': 'process1'})

    composite = Composite(dict(
        processes={'process2': ToyProcess()},
        topology={
            'process2': {
                'port1': ('store_A',),
                'port2': ('store_B',)}}
    ))
    composite.merge(
        processes={process1.name: process1},
        topology={
            process1.name: {
                'port1': ('store_A',),
                'port2': ('store_B',)}})

    expected_topology = {
        'process1': {
            'port1': ('store_A',),
            'port2': ('store_B',)},
        'process2': {
            'port1': ('store_A',),
            'port2': ('store_B',)}}

    assert composite['topology'] == expected_topology


def test_aggregate_composer() -> None:
    config1 = {'name': 'one'}
    aggregate = MetaComposer(
        composers=[ToyComposer(config1)])
    composite1 = aggregate.generate()

    # add composers (list)
    config2 = {
        'name': 'two',
        'process1':  {'name': 'process3'},
        'process2': {'name': 'process4'},
    }
    aggregate.add_composers(
        composers=[ToyComposer(config2)],
        config={'two': {}})
    composite2 = aggregate.generate()

    # add composer (single)
    config3 = {
        'name': 'three',
        'process1':  {'name': 'process5'},
        'process2': {'name': 'process6'},
    }
    aggregate.add_composer(
        composer=ToyComposer(config3),
        config={'three': {}})
    composite3 = aggregate.generate()

    assert all(
        item in composite2['processes'].keys()
        for item in composite1['processes'].keys())
    assert all(
        item in composite3['processes'].keys()
        for item in composite2['processes'].keys())
    assert all(
        item in composite2['topology'].keys()
        for item in composite1['topology'].keys())
    assert all(
        item in composite3['topology'].keys()
        for item in composite2['topology'].keys())
    assert len(composite1['processes']) < len(composite2['processes'])
    assert len(composite2['processes']) < len(composite3['processes'])



def split_divider_int(value, config):
    return [int(value * config['fraction'])] * 2


class ToyDividerProcess(Process):
    defaults = {
        'name': 'divider',
        'x_growth': 1,
        'x_division_fraction': 0.5,
        'x_division_threshold': 10,
    }
    def __init__(self, parameters = None):
        super().__init__(parameters)
        self.agent_id = self.parameters['agent_id']
        self.composer = self.parameters['composer']

    def ports_schema(self):
        return {
            'variable': {
                'x': {
                    '_default': 0,
                    '_emit': True,
                    '_divider': {
                        'divider': split_divider_int,
                        'config': {
                            'fraction': self.parameters['x_division_fraction']}
                    }
                }},
            'agents': {}}

    def next_update(self, timestep, states):
        x = states['variable']['x']
        if x > self.parameters['x_division_threshold']:
            daughter_ids = [
                str(self.agent_id) + '0',
                str(self.agent_id) + '1']
            divide_update = get_divide_update(
                self.composer,
                self.agent_id,
                daughter_ids,
                composer_config={
                    self.parameters['name']: self.parameters},
            )
            return {'agents': divide_update}
        return {'variable': {'x': self.parameters['x_growth']}}


class ToyDividerStep(Step):
    def ports_schema(self):
        return {
            'variable': {
                'x': {
                    '_default': 0,
                },
                '2x': {
                    '_default': 0,
                    '_emit': True,
                    '_updater': 'set',
                },
            },
        }
    def next_update(self, timestep, states):
        x = states['variable']['x']
        return {
            'variable': {
                '2x': 2 * x,
            },
        }


class ToyDivider(Composer):
    defaults: Dict[str, Any] = {
        'divider': {'name': 'divider'}}

    def generate_processes(self, config):
        agent_id = config['agent_id']
        division_config = dict(
            config['divider'],
            agent_id=agent_id,
            composer=self)
        return {
            'divider': ToyDividerProcess(division_config),
        }

    def generate_steps(self, config):
        return {
            'step': ToyDividerStep(),
        }

    def generate_flow(self, config):
        return {
            'step': [],
        }

    def generate_topology(self, config):
        return {
            'divider': {
                'variable': ('variable',),
                'agents': ('..', '..', 'agents'),
            },
            'step': {
                'variable': ('variable',),
            },
        }


if __name__ == '__main__':
    test_composite_initial_state()  # pragma: no cover
    test_composite_parameters()  # pragma: no cover
    test_composite_merge()  # pragma: no cover
    test_get_composite()  # pragma: no cover
    test_aggregate_composer()  # pragma: no cover
    test_override()  # pragma: no cover
    test_composite_initial_state_complex()  # pragma: no cover



"""
=============
Clock Process
=============
"""
from vivarium.core.process import Process

class Clock(Process):
    """ Clock

    A process for tracking global time
    """
    name = 'clock'
    defaults = {
        'time_step': 1.0}

    def initial_state(self, config=None):
        return {}

    def ports_schema(self):
        return {
            'global_time': {
                '_default': 0.0,
                '_updater': 'accumulate'}}

    def next_update(self, timestep, states):
        return {'global_time': timestep}



"""
================
Process Classes
================
"""

import abc
import copy
import cProfile
from multiprocessing import Pipe
from multiprocessing import Process as Multiprocess
from multiprocessing.connection import Connection
import pstats
import pickle
from typing import (
    Any, Dict, Optional, Union, List)
from warnings import warn

from vivarium.library.dict_utils import (
    deep_merge, deep_merge_check, deep_copy_internal)
from vivarium.library.topology import assoc_path, get_in
from vivarium.core.types import (
    HierarchyPath, Schema, State, Update,
    Topology, Flow)

DEFAULT_TIME_STEP = 1.0


def assoc_in(d: dict, path: HierarchyPath, value: Any) -> dict:
    """Insert a value into a dictionary at an arbitrary depth.

    Empty dictionaries will be created as needed to insert the value at
    the specified depth.

    >>> d = {'a': {'b': 1}}
    >>> assoc_in(d, ('a', 'c', 'd'), 2)
    {'a': {'b': 1, 'c': {'d': 2}}}

    Args:
        d: Dictionary to insert into.
        path: Path in the dictionary where the value will be inserted.
            Each element of the path is dictionary key, which will be
            added if not already present. Any given element (except the
            first) refers to a key in the dictionary that is the value
            associated with the immediately preceding path element.
        value: The value to insert.

    Returns:
        Dictionary with the value inserted.
    """
    if path:
        return dict(
            d,
            **{
                path[0]: assoc_in(d.get(path[0], {}), path[1:], value)
            }
        )
    return value


def _override_schemas(
        overrides: Dict[str, Schema],
        processes: Dict[str, 'Process']
) -> None:
    for key, override in overrides.items():
        process = processes[key]
        if isinstance(process, Process):
            process.merge_overrides(override)
        elif isinstance(process, dict):
            _override_schemas(override, process)


def _get_parameters(
        processes: Optional[Dict[str, 'Process']] = None
) -> Dict:
    processes = processes or {}
    parameters: Dict = {}
    for key, value in processes.items():
        if isinstance(value, Process):
            parameters[key] = value.parameters
        elif isinstance(value, dict):
            parameters[key] = _get_parameters(value)
    return parameters


class Process(metaclass=abc.ABCMeta):
    defaults: Dict[str, Any] = {}

    def __init__(self, parameters: Optional[dict] = None) -> None:
        """Process parent class.

        All :term:`process` classes must inherit from this class. Each
        class can provide a ``defaults`` class variable to specify the
        process defaults as a dictionary.

        Note that subclasses should call the superclass init function
        first. This allows the superclass to correctly save the initial
        parameters before they are mutated by subclass constructor code.
        We need access to the original parameters for serialization to
        work properly.

        Args:
            parameters: Override the class defaults. This dictionary may
                also contain the following special keys:

                * ``name``: Saved to ``self.name``.
                * ``_original_parameters``: Returned by
                  ``__getstate__()`` for serialization.
                * ``_no_original_parameters``: If specified with a value
                  of ``True``, original parameters will not be copied
                  during initialization, and ``__getstate__()`` will
                  instead return ``self.parameters``. This puts the
                  responsibility on the user to not mutate process
                  parameters.
                * ``_schema``: Overrides the schema.
                * ``_parallel``: Indicates that the process should be
                  parallelized. ``self.parallel`` will be set to True.
                * ``_condition``: Path to a variable whose value will be
                  returned by
                  :py:meth:`vivarium.core.process.Process.update_condition`.
                * ``time_step``: Returned by
                  :py:meth:`vivarium.core.process.Process.calculate_timestep`.
        """
        parameters = parameters or {}
        if '_original_parameters' in parameters:
            original_parameters = parameters.pop('_original_parameters')
        else:
            original_parameters = parameters
        if parameters.get('_no_original_parameters', False):
            self._original_parameters: Optional[dict] = None
        else:
            try:
                self._original_parameters = copy.deepcopy(
                    original_parameters)
            except TypeError:
                # Copying the parameters failed because some parameters do
                # not support being copied.
                self._original_parameters = None

        if 'name' in parameters:
            self.name = parameters['name']
        elif not hasattr(self, 'name'):
            self.name = self.__class__.__name__

        self.parameters = copy.deepcopy(self.defaults)
        self.parameters = deep_merge(self.parameters, parameters)
        self.schema_override = self.parameters.pop('_schema', {})
        self.parallel = self.parameters.pop('_parallel', False)
        self.condition_path = None

        # set up the conditional state if a condition key is provided
        if '_condition' in self.parameters:
            self.condition_path = self.parameters.pop('_condition')
        if self.condition_path:
            self.merge_overrides(assoc_path({}, self.condition_path, {
                '_default': True,
                '_emit': True,
                '_updater': 'set'}))

        self.parameters.setdefault('time_step', DEFAULT_TIME_STEP)
        self.schema: Optional[dict] = None

    def __getstate__(self) -> dict:
        """Return parameters

        This is sufficient to reproduce the Process if there are no
        hidden states. Processes with hidden states may need to write
        their own __getstate__.

        The original parameters saved by the constructor are used here,
        so any later changes to the parameters will be lost during
        serialization.
        """
        if self.parameters.get('_no_original_parameters', False):
            return self.parameters
        if self._original_parameters is None:
            raise TypeError(
                'Parameters could not be copied, so serialization is '
                'not supported.')
        return self._original_parameters

    def __setstate__(self, parameters: dict) -> None:
        """Initialize process with parameters"""
        self.__init__(parameters)  # type: ignore

    def initial_state(self, config: Optional[dict] = None) -> State:
        """Get initial state in embedded path dictionary.

        Every subclass may override this method.

        Args:
            config (dict): A dictionary of configuration options. All
                subclass implementation must accept this parameter, but
                some may ignore it.

        Returns:
            dict: Subclass implementations must return a dictionary
            mapping state paths to initial values.
        """
        _ = config
        return {}

    def generate_processes(
            self, config: Optional[dict] = None) -> Dict[str, Any]:
        """Do not override this method."""
        if not self.is_step():
            config = config or {}
            name = config.get('name', self.name)
            return {name: self}
        return {}

    def generate_steps(
            self, config: Optional[dict] = None) -> Dict[str, Any]:
        """Do not override this method."""
        if self.is_step():
            config = config or {}
            name = config.get('name', self.name)
            return {name: self}
        return {}

    def generate_topology(self, config: Optional[dict] = None) -> Topology:
        """Do not override this method."""
        config = config or {}
        name = config.get('name', self.name)
        override_topology = config.get('topology', {})
        ports = self.ports()
        return {
            name: {
                port: override_topology.get(port, (port,))
                for port in ports.keys()}}

    def generate_flow(self, config: Optional[dict] = None) -> Flow:
        _ = config
        return {}

    def generate(
            self,
            config: Optional[dict] = None,
            path: HierarchyPath = ()) -> Dict:
        if config is None:
            config = self.parameters
        else:
            default = copy.deepcopy(self.parameters)
            config = deep_merge(default, config)
        config = config or {}
        name = config.get('name', self.name)
        processes = self.generate_processes(config)
        steps = self.generate_steps(config)
        topology = self.generate_topology(config)
        flow = self.generate_flow(config)
        processes_and_steps = deep_copy_internal(processes)
        deep_merge_check(processes_and_steps, steps)
        _override_schemas(
            {name: self.schema_override},
            processes_and_steps)

        # TODO -- this should return a Composite instance,
        # but importing Composite introduces circular imports
        return {
            'processes': assoc_in({}, path, processes),
            'steps': assoc_in({}, path, steps),
            'topology': assoc_in({}, path, topology),
            'flow': assoc_in({}, path, flow),
        }

    def get_schema(self, override: Optional[Schema] = None) -> dict:
        """Get the process's schema, optionally with a schema override.

        Args:
            override: Override schema

        Returns:
            The combined schema.
        """
        ports = copy.deepcopy(self.ports_schema())
        deep_merge(ports, self.schema_override)
        deep_merge(ports, override)
        return ports

    def merge_overrides(self, override: Schema) -> None:
        """Add a schema override to the process's schema overrides.

        Args:
            override: The schema override to add.
        """
        deep_merge(self.schema_override, override)

    def ports(self) -> Dict[str, List[str]]:
        """Get ports and each port's variables.

        Returns:
            A map from port names to lists of the variables that go into
            that port.
        """
        ports_schema = self.ports_schema()
        return {
            port: list(states.keys())
            for port, states in ports_schema.items()}

    def calculate_timestep(self, states: Optional[State]) -> Union[float, int]:
        """Return the next process time step

        A process subclass may override this method to implement
        adaptive timesteps. By default it returns self.parameters['time_step'].
        """
        _ = states
        return self.parameters['time_step']

    def default_state(self) -> State:
        """Get the default values of the variables in each port.

        The default values are computed based on the schema.

        Returns:
            A state dictionary that assigns each variable's default
            value to that variable.
        """
        schema = self.get_schema()
        state: State = {}
        for port, states in schema.items():
            for key, value in states.items():
                if '_default' in value:
                    if port not in state:
                        state[port] = {}
                    state[port][key] = value['_default']
        return state

    def is_deriver(self) -> bool:
        """Check whether this process is a :term:`deriver`.

        .. deprecated:: 0.3.14
           Derivers have been deprecated in favor of :term:`steps`, so
           please override ``is_step`` instead of ``is_deriver``.
           Support for Derivers may be removed in a future release.

        Returns:
            Whether this process is a deriver. This class always returns
            ``False``, but subclasses may change this.
        """
        return False

    def is_step(self) -> bool:
        """Check whether this process is a :term:`step`.

        Returns:
            Whether this process is a step. This class always returns
            ``False``, but subclasses may change this behavior.
        """
        method = getattr(self.is_deriver, '__func__', None)
        if method and method is not Process.is_deriver:
            # `self` is an instance of a subclass that has overridden
            # `is_deriver`.
            warn(
                'is_deriver() is deprecated. Use is_step() instead.',
                category=FutureWarning)
            return self.is_deriver()
        return False

    def get_private_state(self) -> State:
        """Get the process's private state.

        Processes can store state in instance variables instead of in
        the :term:`stores` that hold the simulation-wide state. These
        instance variables hold a private state that is not shared with
        other processes. You can override this function to let
        :term:`experiments` emit the process's private state.

        Returns:
            An empty dictionary. You may override this behavior to
            return your process's private state.
        """
        return {}  # pragma: no cover

    @abc.abstractmethod
    def ports_schema(self) -> Schema:
        """Get the schemas for each port.

        This must be overridden by any subclasses.

        Use the glob '*' schema to declare expected sub-store structure,
        and view all child values of the store:

          .. code-block:: python

            schema = {
                'port1': {
                    '*': {
                        '_default': 1.0
                    }
                }
            }

        Use the glob '**' schema to connect to an entire sub-branch, including
        child nodes, grandchild nodes, etc:

          .. code-block:: python

            schema = {
                'port1': '**'
            }

        Ports flagged as output-only won't be viewed through the next_update's
        states, which can save some overhead time:

          .. code-block:: python

            schema = {
              'port1': {
                 '_output': True,
                 'A': {'_default': 1.0},
              }
            }

        Returns:
            A dictionary that declares which states are expected by the
            processes, and how each state will behave. State keys can be
            assigned properties through schema_keys declared in
            :py:class:`vivarium.core.store.Store`.
        """
        return {}  # pragma: no cover

    @abc.abstractmethod
    def next_update(
            self, timestep: Union[float, int], states: State) -> Update:
        """Compute the next update to the simulation state.

        Args:
            timestep: The duration for which the update should be
                computed.
            states: The pre-update simulation state. This will take the
                same form as the process's schema, except with a value
                for each variable.

        Returns:
            An empty dictionary for now. This should be overridden by
            each subclass to return an update.
        """
        return {}  # pragma: no cover

    def update_condition(
            self, timestep: Union[float, int], states: State) -> Any:
        """Determine whether this process runs.

        Args:
            timestep: The duration for which an update.
            states: The pre-update simulation state.

        Returns:
            Boolean for whether this process runs. True by default.
        """

        _ = timestep
        _ = states

        # use the given condition key if it was provided
        if self.condition_path:
            return get_in(states, self.condition_path)

        return True


class Step(Process, metaclass=abc.ABCMeta):
    """Base class for steps."""

    def __init__(self, parameters: Optional[dict] = None) -> None:
        parameters = parameters or {}
        super().__init__(parameters)

    def is_step(self) -> bool:
        """Returns ``True`` to signal that this process is a step."""
        return True


#: Deriver is just an alias for :py:class:`Step` now that Derivers have
#: been deprecated.
Deriver = Step


def _run_update(
        connection: Connection, process: Process,
        profile: bool) -> None:
    if profile:
        profiler = cProfile.Profile()
        profiler.enable()
    running = True

    while running:
        interval, states = connection.recv()

        # stop process by sending -1 as the interval
        if interval == -1:
            running = False

        else:
            update = process.next_update(interval, states)
            connection.send(update)

    if profile:
        profiler.disable()
        stats = pstats.Stats(profiler)
        connection.send(stats.stats)  # type: ignore

    connection.close()


class ParallelProcess:
    def __init__(self, process: Process, profile: bool = False) -> None:
        """Wraps a :py:class:`Process` for multiprocessing.

        To run a simulation distributed across multiple processors, we
        use Python's multiprocessing tools. This object runs in the main
        process and manages communication between the main (parent)
        process and the child process with the :py:class:`Process` that
        this object manages.

        Args:
            process: The Process to manage.
            profile: Whether to use cProfile to profile the subprocess.
        """
        self.process = process
        self.profile = profile
        self.stats: Optional[pstats.Stats] = None
        self.parent, self.child = Pipe()
        self.multiprocess = Multiprocess(
            target=_run_update,
            args=(self.child, self.process, self.profile))
        self.multiprocess.start()

    def update(
            self, interval: Union[float, int], states: State) -> None:
        """Request an update from the process.

        Args:
            interval: The length of the timestep for which the update
                should be computed.
            states: The pre-update state of the simulation.
        """
        self.parent.send((interval, states))

    def get(self) -> Update:
        """Get an update from the process.

        Returns:
            The update from the process.
        """
        return self.parent.recv()

    def end(self) -> None:
        """End the child process.

        If profiling was enabled, then when the child process ends, it
        will compile its profiling stats and send those to the parent.
        The parent then saves those stats in ``self.stats``.
        """
        self.parent.send((-1, None))
        if self.profile:
            self.stats = pstats.Stats()
            self.stats.stats = self.parent.recv()  # type: ignore
        self.multiprocess.join()


class ToySerializedProcess(Process):

    defaults: Dict[str, list] = {
        'list': [],
    }

    def __init__(self, parameters: Optional[dict] = None) -> None:
        super().__init__(parameters)
        self.parameters['list'].append(1)

    def ports_schema(self) -> Schema:
        return {}

    def next_update(self, timestep: float, states: State) -> Update:
        return {}


class ToySerializedProcessInheritance(Process):

    defaults = {
        '1': 1,
    }

    def __init__(self, parameters: Optional[dict] = None) -> None:
        parameters = parameters or {}
        super().__init__({
            '2': parameters['1'],
            '_original_parameters': parameters,
        })

    def ports_schema(self) -> Schema:
        return {}

    def next_update(self, timestep: float, states: State) -> Update:
        return {}


def test_serialize_process() -> None:
    proc = ToySerializedProcess()
    proc_pickle = pickle.loads(pickle.dumps(proc))

    assert proc.parameters['list'] == [1]
    # If we pickled using `self.parameters` instead of
    # `self._original_parameters`, this list would be [1, 1].
    assert proc_pickle.parameters['list'] == [1]


def test_serialize_process_inheritance() -> None:
    a = ToySerializedProcessInheritance({'1': 0})
    a2 = pickle.loads(pickle.dumps(a))
    assert a2.parameters['2'] == 0



import time

from vivarium.core.engine import Engine
from vivarium.core.process import Process
from vivarium.core.types import Schema, State, Update


class ProcessA(Process):

    defaults = {
        'sleep': 0.2,
    }

    def ports_schema(self) -> Schema:
        return {}

    def next_update(self, timestep: float, states: State) -> Update:
        time.sleep(self.parameters['sleep'])
        return {}


class ProcessB(Process):

    defaults = {
        'sleep': 0.1,
        '_parallel': True,
    }

    def ports_schema(self) -> Schema:
        return {}

    def next_update(self, timestep: float, states: State) -> Update:
        time.sleep(self.parameters['sleep'])
        return {}


def test_profiler() -> None:
    engine = Engine(
        processes={
            'processA': ProcessA(),
            'processB': ProcessB(),
        },
        topology={
            'processA': {},
            'processB': {},
        },
        profile=True,
    )
    engine.update(3)
    engine.end()
    assert engine.stats is not None
    stats = engine.stats.strip_dirs()
    process_a_runtime = stats.stats[  # type: ignore
        ('test_profiler.py', 17, 'next_update')][3]
    process_b_runtime = stats.stats[  # type: ignore
        ('test_profiler.py', 32, 'next_update')][3]

    assert 0.6 <= process_a_runtime <= 0.7
    assert 0.3 <= process_b_runtime <= 0.4



from vivarium.composites.toys import ExchangeA
from vivarium.core.composer import Composite



def test_multi_composite() -> None:

    composite1 = Composite()

    processes = {
        'A': ExchangeA()
    }
    topology = {
        'A': {
            'internal': ('in',),
            'external': ('out',)}}
    composite2 = Composite({
            'processes': processes,
            'topology': topology,
        })

    composite3 = Composite()
    assert not composite3['processes']
    assert not composite3['topology']
    print(f"composite3 processes: {composite3['processes']}")
    print(f"composite3 topology: {composite3['topology']}")

    composite1.merge(composite=composite2, path=('agents', '1',))

    composite4 = Composite()
    assert not composite4['processes']
    assert not composite4['topology']
    print(f"composite4 processes: {composite4['processes']}")
    print(f"composite4 topology: {composite4['topology']}")





if __name__ == '__main__':
    test_multi_composite()



from typing import Any, Dict
import logging as log

from vivarium.core.process import Deriver


def daughter_phylogeny_id(mother_id):
    return [
        str(mother_id) + '0',
        str(mother_id) + '1']


def pass_threshold(value, config):
    threshold = config['threshold']
    if value >= threshold:
        return True
    return False


def get_divide_update(
        daughter_composer, mother_id, daughter_ids, composer_config=None):
    composer_config = composer_config or {}
    daughter_updates = []
    for daughter_id in daughter_ids:
        composer = daughter_composer.generate({
            'agent_id': daughter_id,
            **composer_config})
        daughter_updates.append({
            'key': daughter_id,
            'processes': composer['processes'],
            'steps': composer['steps'],
            'flow': composer['flow'],
            'topology': composer['topology'],
            'initial_state': {}})
    return {
        '_divide': {
            'mother': mother_id,
            'daughters': daughter_updates}}


class Division(Deriver):
    """ Division Process """
    defaults: Dict[str, Any] = {
        'daughter_ids_function': daughter_phylogeny_id,
        'condition_function': pass_threshold,
        'condition_config': {
            'threshold': None
        }
    }

    def __init__(self, parameters=None):
        super().__init__(parameters)
        self.condition_function = self.parameters['condition_function']
        self.condition_config = self.parameters['condition_config']

        # must provide a composer to generate new daughters
        self.agent_id = self.parameters['agent_id']
        self.composer = self.parameters['composer']
        self.daughter_ids_function = self.parameters['daughter_ids_function']

    def ports_schema(self):
        return {
            'variable': {},
            'agents': {}}

    def next_update(self, timestep, states):
        variable = states['variable']
        if self.condition_function(variable, self.condition_config):
            daughter_ids = self.daughter_ids_function(self.agent_id)
            divide_update = get_divide_update(
                self.composer, self.agent_id, daughter_ids)
            log.info(
                'DIVIDE! \n--> MOTHER: %s \n--> DAUGHTERS: %s',
                str(self.agent_id), str(daughter_ids))
            return {'agents': divide_update}
        return {}



"""
=====
Types
=====
"""

from typing import Tuple, Dict, Any, Union, Sequence


#: Relative path between nodes in the :term:`hierarchy`. Like Unix file
#: paths, ".." refers to the parent directory.
HierarchyPath = Tuple[str, ...]

#: Mapping from :term:`ports` to paths that specify which node in the
#: :term:`hierarchy` should be wired to each port.
Topology = Dict[str, Union[HierarchyPath, dict, object]]

#: Mapping from processes names to Processes, which can be embedded in
#: a hierarchy.
Processes = Dict[str, Any]

#: Mapping from step names to Steps, which can be embedded in a
#: hierarchy.
Steps = Dict[str, Any]

#: Mapping from step names to sequences of HierarchyPaths that specify
#: the step's dependencies.
Flow = Dict[str, Sequence[HierarchyPath]]

#: A dictionary that specifies a :term:`schema`.
Schema = Dict[str, Any]

#: A dictionary that has the form of a :term:`schema`, except instead of
#: specifying the properties of each :term:`variable`, it specifies each
#: variable's value.
State = Dict[str, Any]

#: A dictionary defining an :term:`update`.
Update = Dict[str, Any]

#: A dictionary that specifies the :term:`processes` and
#: :term:`topology` of a :term:`composite`.
CompositeDict = Dict[str, Any]
# TODO(jerry): ^ Dict values should be Union[Process, Topology, CompositeDict]
#  but Process would make recursive imports and CompositeDict would make
#  recursive types. Fix this by switching from a dict to a class or dataclass.

#: A dictionary that contains the retrieved output of an :term:`experiment`
OutputDict = Dict[Union[Tuple, str], Any]



"""
======================================
Toy Glucose Phosphorylation Experiment
======================================

This is a toy example referred to by the documentation.
"""

import os

from vivarium.composites.injected_glc_phosphorylation import (
    InjectedGlcPhosphorylation,
)
from vivarium.core.composition import (
    simulate_experiment,
    EXPERIMENT_OUT_DIR,
)
from vivarium.plots.simulation_output import plot_simulation_output
from vivarium.core.emitter import (
    path_timeseries_from_data,
    timeseries_from_data,
)
from vivarium.core.engine import Engine


NAME = 'glucose_phosphorylation'
OUT_DIR = os.path.join(EXPERIMENT_OUT_DIR, NAME)


def glucose_phosphorylation_experiment(config=None):
    if config is None:
        config = {}
    default_config = {
        'injected_glc_phosphorylation': {},
        'emitter': {
            'type': 'timeseries',
        },
        'initial_state': {},
    }
    default_config.update(config)
    config = default_config
    compartment = InjectedGlcPhosphorylation(
        config['injected_glc_phosphorylation'])
    compartment_dict = compartment.generate()
    experiment = Engine(**{
        'processes': compartment_dict['processes'],
        'topology': compartment_dict['topology'],
        'emitter': config['emitter'],
        'initial_state': config['initial_state'],
    })
    return experiment


def run_experiment():
    experiment = glucose_phosphorylation_experiment()
    settings = {
        'timestep': 1,
        'return_raw_data': True,
    }
    data = simulate_experiment(experiment, settings)
    experiment.end()
    return data


def test_experiment():
    data = run_experiment()
    path_ts = path_timeseries_from_data(data)
    # At every timestep, the changes in ADP and G6P should be equal, and
    # the change in GLC should be the same but with the opposite sign.
    atp = path_ts[('cell', 'ATP')]
    adp = path_ts[('cell', 'ADP')]
    g6p = path_ts[('cell', 'G6P')]
    glc = path_ts[('cell', 'GLC')]

    assert len(atp) == len(adp) == len(g6p) == len(glc)

    for i in range(len(atp) - 1):
        delta_adp = adp[i + 1] - adp[i]
        delta_g6p = g6p[i + 1] - g6p[i]
        delta_glc = glc[i + 1] - glc[i]
        delta_atp = atp[i + 1] - atp[i]

        assert delta_adp == delta_g6p, 'index: {}'.format(i)
        assert delta_atp > 0, 'index: {}'.format(i)
        assert delta_glc < 0, 'index: {}'.format(i)

    print(path_ts)


def main():
    test_experiment()
    data = run_experiment()
    agents_plot_settings = {
        'agents_key': 'agents',
    }
    plot_simulation_output(
        timeseries_from_data(data),
        agents_plot_settings,
        OUT_DIR,
        'simulation',
    )


if __name__ == '__main__':
    main()



"""
===============================================
Registry of Updaters, Dividers, and Serializers
===============================================

You should interpret words and phrases that appear fully capitalized in
this document as described in :rfc:`2119`. Here is a brief summary of
the RFC:

* "MUST" indicates absolute requirements. Vivarium may not work
  correctly if you don't follow these.
* "SHOULD" indicates strong suggestions. You might have a valid reason
  for deviating from them, but be careful that you understand the
  ramifications.
* "MAY" indicates truly optional features that you can include or
  exclude as you wish.

--------
Updaters
--------

Each :term:`updater` is defined as a function whose name begins with
``update_``. Vivarium uses these functions to apply :term:`updates` to
:term:`variables`. Updater names are registered in
:py:data:`updater_registry`, which maps these names to updater functions.

Updater API
===========

An updater function SHOULD have a name that begins with ``update_``. The
function MUST accept exactly three positional arguments: the first MUST
be the current value of the variable (i.e. before applying the update),
the second MUST be the value associated with the variable in the update,
and the third MUST be either a dictionary of states from the simulation
hierarchy or None if no ``port_mapping`` key was specified in the
updater definition. The function SHOULD not accept any other parameters.
The function MUST return the updated value of the variable only.

--------
Dividers
--------

Each :term:`divider` is defined by a function that follows the API we
describe below. Vivarium uses these dividers to generate daughter cell
states from the mother cell's state. Divider names are registered in
:py:data:`divider_registry`, which maps these names to divider functions.

Divider API
===========

Each divider function SHOULD have a name prefixed with ``divide_``. The
function MUST accept a single positional argument, the value of the
variable in the mother cell. It SHOULD accept no other arguments. The
function MUST return either:

1. A :py:class:`list` with two elements: the values of the variables in
   each of the daughter cells.
2. ``None``, in which case division will be skipped for that variable.

.. note:: Dividers MAY not be deterministic and MAY not be symmetric.
    For example, a divider splitting an odd, integer-valued value may
    randomly decide which daughter cell receives the remainder.
"""
import copy
import random
import re

import numpy as np

from vivarium.library.dict_utils import deep_merge
from vivarium.library.units import Quantity


class Registry(object):
    def __init__(self):
        """A Registry holds a collection of functions or objects."""
        self.registry = {}
        self.main_keys = []

    def register(self, key, item, alternate_keys=tuple()):
        """Add an item to the registry.

        Args:
            key: Item key.
            item: The item to add.
            alternate_keys: Additional keys under which to register the
                item. These keys will not be included in the list
                returned by ``Registry.list()``.

                This may be useful if you want to be able to look up an
                item in the registry under multiple keys. For example,
                in a registry of serializers, you could register a
                serializer with its name as ``key`` and the types it can
                serialize under ``alternate_keys``. Then, when presented
                with an object to serialize, you could use the object's
                type to look up the appropriate serializer.
        """
        keys = [key]
        keys.extend(alternate_keys)
        for registry_key in keys:
            if registry_key in self.registry:
                if item != self.registry[registry_key]:
                    raise Exception('registry already contains an entry for {}: {} --> {}'.format(
                        registry_key, self.registry[key], item))
            else:
                self.registry[registry_key] = item
        self.main_keys.append(key)

    def access(self, key):
        """Get an item by key from the registry."""
        return self.registry.get(key)

    def list(self):
        return list(self.main_keys)


# Initialize registries
# These are imported into module __init__.py files,
# where the functions and classes are registered upon import

#: Maps process names to :term:`process classes`
process_registry = Registry()

#: Map updater names to :term:`updater` functions
updater_registry = Registry()

#: Map divider names to :term:`divider` functions
divider_registry = Registry()

#: Map serializer names to :term:`serializer` classes
serializer_registry = Registry()

#: Map serializer names to :term:`Emitter` classes
emitter_registry = Registry()


# Updaters, Dividers, and Serializers
# These can be defined here, but are registered in the base module's __init__.py file

# Updater functions
def update_merge(current_value, new_value):
    """Merge Updater

    Returns:
        dict: The merger of ``current_value`` and ``new_value``. For any
        shared keys, the value in ``new_value`` is used.
    """
    update = current_value.copy()
    for k, v in current_value.items():
        new = new_value.get(k)
        if isinstance(new, dict):
            update[k] = deep_merge(copy.deepcopy(v), new)
        else:
            update[k] = new
    return update


def update_set(current_value, new_value):
    """Set Updater

    Returns:
        The value provided in ``new_value``.
    """
    return new_value


def update_null(current_value, new_value):
    """Null Updater

    Returns:
        The value provided in ``current_value``.
    """
    return current_value


def update_accumulate(current_value, new_value):
    """Accumulate Updater

    Returns:
        The sum of ``current_value`` and ``new_value``.
    """
    return current_value + new_value


def update_nonnegative_accumulate(current_value, new_value):
    """Non-negative Accumulate Updater

    Returns:
        The sum of ``current_value`` and ``new_value`` if positive, 0 if negative.
    """
    updated_value = current_value + new_value
    if isinstance(updated_value, np.ndarray):
        updated_value[updated_value < 0] = 0
        return updated_value
    elif updated_value >= 0:
        return updated_value
    else:
        return 0 * updated_value


def update_dictionary(current, update):
    """Dictionary Updater
    Updater that translates _add and _delete -style updates
    into operations on a dictionary.

    Expects current to be a dictionary, with no restriction on the types of objects
    stored within it, and no defaults values.
    """
    result = current

    for key, value in update.items():
        if key == "_add":
            for added_value in value:
                added_key = added_value["key"]
                added_state = added_value["state"]
                result[added_key] = added_state
        elif key == "_delete":
            for k in value:
                del result[k]
        elif key in result:
            result[key].update(value)
        else:
            raise Exception(f"Invalid dict_value_updater key: {key}")
    return result


# Divider functions
def divide_set(state):
    """Set Divider

    Returns:
        A list ``[state, state]``. No copying is performed.
    """
    return [state, state]


def divide_set_value(state, config):
    """Set Value Divider
    Args:
        'state': value
    Returns:
        A list ``[value, value]``. No copying is performed.
    """
    value = config['value']
    return [value, value]


def divide_split(state):
    """Split Divider

    Arguments:
        state: Must be an :py:class:`int`, a :py:class:`float`, or a
            :py:class:`str` of value ``Infinity``.

    Returns:
        A list, each of whose elements contains half of ``state``. If
        ``state`` is an :py:class:`int`, the remainder is placed at
        random in one of the two elements. If ``state`` is infinite, the
        return value is ``[state, state]`` (no copying is done).

    Raises:
        Exception: if ``state`` is of an unrecognized type.
    """
    if isinstance(state, (int, np.integer)):
        remainder = state % 2
        half = int(state / 2)
        if random.choice([True, False]):
            return [half + remainder, half]
        else:
            return [half, half + remainder]
    elif state == float('inf') or state == 'Infinity':
        # some concentrations are considered infinite in the environment
        # an alternative option is to not divide the local environment state
        return [state, state]
    elif isinstance(state, (float, Quantity)):
        half = state/2
        return [half, half]
    else:
        raise Exception('can not divide state {} of type {}'.format(state, type(state)))


def divide_binomial(state):
    """Binomial Divider
    """
    counts_1 = np.random.binomial(state, 0.5)
    counts_2 = state - counts_1
    return [counts_1, counts_2]


def divide_zero(state):
    """Zero Divider

    Returns:
        ``[0, 0]`` regardless of input
    """
    return [0, 0]


def divide_split_dict(state):
    """Split-Dictionary Divider

    Returns:
        A list of two dictionaries. The first dictionary stores the
        first half of the key-value pairs in ``state``, and the second
        dictionary stores the rest of the key-value pairs.

        .. note:: Since dictionaries are unordered, you should avoid
            making any assumptions about which keys will be sent to
            which daughter cell.
    """
    if state is None:
        state = {}
    d1 = dict(list(state.items())[len(state) // 2:])
    d2 = dict(list(state.items())[:len(state) // 2])
    return [d1, d2]


def assert_no_divide(state):
    """Assert that the variable is never divided

    Raises:
        AssertionError: If the variable is divided
    """
    raise AssertionError('Variable cannot be divided')


def divide_null(state):
    """Divider that causes the variable to be skipped during division.

    Returns:
        ``None`` so that no divided values are provided to the daughter
        cells. This is useful for process objects, which are handled
        separately during division.
    """
    return None


# Serializers
class Serializer:
    """Base serializer class.

    Serializers work together to convert Python objects, which may be
    collections of many different kinds of objects, into JSON-compatible
    representations. Those representations can then be deserialized to
    recover the original object.

    Most serializers convert an object to a string. To implement a
    serializer of this type, you only need to implement the
    :py:meth:`vivarium.core.registry.Serializer.serialize_to_string` and
    :py:meth:`vivarium.core.registry.Serializer.deserialize()` methods.

    Args:
        name: Name of the serializer. Defaults to the class name.
        exclusive_types: Types that are exclusively handled by this
            serializer. These can be used to quickly lookup which
            serializer handles a particular type. If an object does not
            match any of these exclusive types, the `can_serialize`
            method will be used as a fallback.

            Note that exclusive type matching is exact--inheritance is
            not accounted for. This means that if your serializer has an
            exclusive type of A and B inherits from A, an instance of B
            will not match to your serializer's exclusive types.
    """
    REGEX_FOR_NAME = re.compile('[A-Za-z0-9-_]+')
    REGEX_FOR_SERIALIZED_ANY_TYPE = re.compile(
        f'!{REGEX_FOR_NAME.pattern}\\[(.*)\\]')

    def __init__(self, name='', exclusive_types=tuple()):
        self.name = name or self.__class__.__name__
        self.regex_for_serialized = re.compile(f'!{self.name}\\[(.*)\\]')
        self.alternate_keys = (
            str(exclusive_type)
            for exclusive_type in exclusive_types
        )

    def serialize_to_string(self, data):
        """Serialize some data.

        Subclasses should override this function,

        Args:
            data: Data to serialize.

        Returns:
            The serialized data.
        """
        raise NotImplementedError(
            f'{self} has not implemented serialize_to_string().')

    def serialize(self, data):
        """Serialize some data.

        Subclasses that are serializing to strings can just implement
        :py:meth:`vivarium.core.registry.Serializer.serialize_to_string`,
        and this method will use that implementation to generate a
        string that serializes the data.

        Subclasses with more complex serialization needs must override
        this function.

        Args:
            data: Data to serialize.

        Returns:
            The serialized data.
        """
        string_serialization = self.serialize_to_string(data)
        if not isinstance(string_serialization, str):
            raise ValueError(
                f'{self}.serialize_to_string() returned invalid '
                f'serialization: {string_serialization}')

        return f'!{self.name}[{string_serialization}]'

    def can_serialize(self, data):
        """Check whether the serializer can serialize some data.

        Every subclass must implement this method.
        """
        raise NotImplementedError(
            f'{self} has not implemented can_serialize().')

    def deserialize_from_string(self, data):
        """Deserialize data from a string.

        This function is the inverse of
        :py:meth:`vivarium.core.registry.Serializer.serialize_to_string`.
        It need only be overridden in a subclass if that subclass does
        not override
        :py:meth:`vivarium.core.registry.Serializer.deserialize`.
        """
        raise NotImplementedError(
            f'{self} has not implemented deserialize_from_string().')

    def deserialize(self, data):
        """Deserialize some data.

        Subclasses relying on
        :py:meth:`vivarium.core.registry.Serializer.serialize_to_string`
        should not override this function and instead override
        :py:meth:`vivarium.core.registry.Serializer.deserialize_from_string`.

        Args:
            data: Serialized data to deserialize.

        Returns:
            The deserialized data.
        """
        string_serialization = self.regex_for_serialized.fullmatch(
            data).group(1)
        return self.deserialize_from_string(string_serialization)

    def can_deserialize(self, data):
        """Check whether the serializer can deserialize an object.

        By default, this function checks that ``data`` is a string of the
        form produced by the default implementation of
        :py:meth:`vivarium.core.registry.Serializer.serialize`. If a
        subclass overrides this ``serialize()`` method, it should also
        override ``can_deserialize()``.
        """
        if not isinstance(data, str):
            return False
        return bool(self.regex_for_serialized.fullmatch(data))



'''
==================
Topology Utilities
==================
'''

import copy
import re

from vivarium.library.dict_utils import deep_merge, deep_merge_multi_update


def get_in(d, path, default=None):
    '''Get the value from a dictionary by its path.

    >>> d = {'a': {'b': 'c', 'd': 'e'}}
    >>> get_in(d, ('a', 'b'))
    'c'
    >>> get_in(d, ('a', 'z'))
    >>> get_in(d, ('a', 'z'), 'y')
    'y'
    '''
    if path:
        head = path[0]
        if head in d:
            return get_in(d[head], path[1:], default)
        return default
    return d


def delete_in(d, path):
    '''Delete an item from a dictionary by its path.

    >>> d = {'a': {'b': 'c', 'd': 'e'}}
    >>> delete_in(d, ('a', 'b'))
    >>> d
    {'a': {'d': 'e'}}
    '''
    if len(path) > 0:
        head = path[0]
        if len(path) == 1:
            # at the node to be deleted
            if head in d:
                del d[head]
        elif head in d:
            # down = d[head]
            delete_in(d[head], path[1:])


def assoc_path(d, path, value):
    '''Insert ``value`` into the dictionary ``d`` at ``path``.

    >>> d = {'a': {'b': 'c'}}
    >>> assoc_path(d, ('a', 'd'), 'e')
    {'a': {'b': 'c', 'd': 'e'}}
    >>> d
    {'a': {'b': 'c', 'd': 'e'}}

    Create new dictionaries recursively as needed.
    '''

    if path:
        head = path[0]
        if len(path) == 1:
            d[head] = value
        else:
            if head not in d:
                d[head] = {}
            assoc_path(d[head], path[1:], value)
    elif isinstance(value, dict):
        deep_merge(d, value)
    return d


def without(d, removing):
    '''Get a copy of ``d`` without the key specified by ``removing``.'''
    return {
        key: value
        for key, value in d.items()
        if key != removing}


def update_in(d, path, f):
    '''Update every value in a dictionary based on ``f``.

    Args:
        d: The dictionary ``path`` applies to. This object is not
            modified.
        path: Path to the sub-dictionary within ``d`` that should be
            updated.
        f: Function to call on every value in the dictionary to update.
            The updated dictionary's values will be return values from
            ``f``.

    Returns:
        A copy of ``d`` with all the values under ``path`` updated to
        the value returned when ``f`` is called on the original value.
    '''
    if path:
        head = path[0]
        d.setdefault(head, {})
        # updated = copy.deepcopy(d)
        updated = copy.copy(d)
        updated[head] = update_in(d[head], path[1:], f)
        return updated
    return f(d)


def paths_to_dict(path_list, f=lambda x: x):
    '''Create a new dictionary that has the paths in ``path_list``.

    Args:
        path_list: A list of tuples ``(path, value)``.
        f: A function to apply to each value before inserting it into
            the dictionary.

    Returns:
        A new dictionary with the specified values (after being passed
        through ``f``) at each associated path.
    '''
    d = {}
    for path, node in path_list:
        assoc_path(d, path, f(node))
    return d


def dict_to_paths(root, d):
    """Get all the paths in a dictionary.

    For example:

    >>> root = ('root', 'subroot')
    >>> d = {
    ...     'a': {
    ...         'b': 'c',
    ...     },
    ...     'd': 'e',
    ... }
    >>> dict_to_paths(root, d)
    [(('root', 'subroot', 'a', 'b'), 'c'), (('root', 'subroot', 'd'), 'e')]
    """
    if isinstance(d, dict):
        deeper = []
        for key, down in d.items():
            paths = dict_to_paths(root + (key,), down)
            deeper.extend(paths)
        return deeper
    else:
        return [(root, d)]


def inverse_topology(outer, update, topology, inverse=None):
    '''
    Transform an update from the form its process produced into
    one aligned to the given topology.

    The inverse of this function (using a topology to construct a view for
    the perspective of a Process ports_schema()) lives in `Store`, called
    `topology_state`. This one stands alone as it does not require a store
    to calculate.
    '''

    inverse = inverse or {}

    for key, path in topology.items():
        if key == '*':
            if isinstance(path, dict):
                node = inverse
                if '_path' in path:
                    inner = normalize_path(outer + path['_path'])
                    path = without(path, '_path')
                else:
                    inner = outer

                for child, child_update in update.items():
                    inverse = inverse_topology(
                        inner + (child,),
                        update[child],
                        path,
                        inverse)
            else:
                for child, child_update in update.items():
                    inner = normalize_path(outer + path + (child,))
                    if isinstance(child_update, dict):
                        inverse = update_in(
                            inverse,
                            inner,
                            lambda current: deep_merge(
                                current, child_update))
                    else:
                        assoc_path(inverse, inner, child_update)

        elif key in update:
            value = update[key]
            if isinstance(path, dict):
                if '_path' in path:
                    inner = normalize_path(outer + path['_path'])
                    path = without(path, '_path')

                    for update_key in update[key].keys():
                        if update_key not in path and '*' not in path:
                            path[update_key] = (update_key,)
                else:
                    inner = outer

                inverse = inverse_topology(
                    inner,
                    value,
                    path,
                    inverse)
            else:
                inner = normalize_path(outer + path)
                if isinstance(value, dict):
                    inverse = update_in(
                        inverse,
                        inner,
                        lambda current: deep_merge_multi_update(current, value))
                else:
                    assoc_path(inverse, inner, value)
    return inverse


def normalize_path(path):
    """Make a path absolute by resolving ``..`` elements."""
    progress = []
    for step in path:
        if step == '..' and len(progress) > 0:
            progress = progress[:-1]
        else:
            progress.append(step)
    return tuple(progress)


def convert_path_style(path):
    if isinstance(path, str):
        path = re.sub(r'<', '..<', path)
        path = tuple(re.split('<|>', path))

    return path


class TestUpdateIn:
    d = {
        'foo': {
            1: {
                'a': 'b',
            },
        },
        'bar': {
            'c': 'd',
        },
    }

    def test_simple(self):
        updated = copy.deepcopy(self.d)
        updated = update_in(
            updated, ('foo', 1, 'a'), lambda current: 'updated')
        expected = {
            'foo': {
                1: {
                    'a': 'updated',
                },
            },
            'bar': {
                'c': 'd',
            },
        }
        assert updated == expected

    def test_add_leaf(self):
        updated = copy.deepcopy(self.d)
        updated = update_in(
            updated, ('foo', 1, 'new'), lambda current: 'updated')
        expected = {
            'foo': {
                1: {
                    'a': 'b',
                    'new': 'updated',
                },
            },
            'bar': {
                'c': 'd',
            },
        }
        assert updated == expected

    def test_add_dict(self):
        updated = copy.deepcopy(self.d)
        updated = update_in(
            updated, ('foo', 2), lambda current: {'a': 'updated'})
        expected = {
            'foo': {
                1: {
                    'a': 'b',
                },
                2: {
                    'a': 'updated',
                },
            },
            'bar': {
                'c': 'd',
            },
        }
        assert updated == expected

    def test_complex_merge(self):
        updated = copy.deepcopy(self.d)
        updated = update_in(
            updated, ('foo',),
            lambda current: deep_merge(
                current,
                {'foo': {'a': 'updated'}, 'b': 2}),
            )
        expected = {
            'foo': {
                'foo': {
                    'a': 'updated',
                },
                'b': 2,
                1: {
                    'a': 'b',
                },
            },
            'bar': {
                'c': 'd',
            },
        }
        assert updated == expected

    def test_add_to_root(self):
        updated = copy.deepcopy(self.d)
        updated = update_in(
            updated,
            tuple(),
            lambda current: deep_merge(current, ({'a': 'updated'})),
        )
        expected = {
            'foo': {
                1: {
                    'a': 'b',
                },
            },
            'bar': {
                'c': 'd',
            },
            'a': 'updated'
        }
        assert updated == expected

    def test_set_root(self):
        updated = copy.deepcopy(self.d)
        updated = update_in(
            updated, tuple(), lambda current: {'a': 'updated'})
        expected = {
            'a': 'updated',
        }
        assert updated == expected


def test_inverse_topology():
    update = {
        'port1': {
            'a': 5},
        'port2': {
            'b': 10},
        'port3': {
            'b': 10},
        'global': {
            'c': 20}}

    topology = {
        'port1': ('boundary', 'x'),
        'global': ('boundary',),
        'port2': ('boundary', 'y'),
        'port3': ('boundary', 'x')}

    path = ('agent',)
    inverse = inverse_topology(path, update, topology)
    expected_inverse = {
        'agent': {
            'boundary': {
                'x': {
                    'a': 5,
                    'b': 10},
                'y': {
                    'b': 10},
                'c': 20}}}

    assert inverse == expected_inverse


def test_deletion():
    nested = {
        'A': {
            'AA': 5,
            'AB': {
                'ABC': 11}},
        'B': {
            'BA': 6}}

    delete_in(nested, ('A', 'AA'))
    assert 'AA' not in nested['A']


def test_in():
    blank = {}
    path = ['where', 'are', 'we']
    assoc_path(blank, path, 5)
    print(blank)
    print(get_in(blank, path))
    blank = update_in(blank, path, lambda x: x + 6)
    print(blank)

def test_path_declare():

    path_down = 'path>to>store'
    new_path_down = convert_path_style(path_down)
    assert new_path_down == ('path', 'to', 'store')

    path_up = '<<store'
    new_path_up = convert_path_style(path_up)
    assert new_path_up == ('..', '..', 'store')


if __name__ == '__main__':
    test_path_declare()



import math
import re
from typing import Any

import numpy as np

from vivarium.core.process import Process
from vivarium.core.serialize import serialize_value, deserialize_value
from vivarium.core.registry import Serializer
from vivarium.library.units import units


class SerializeProcess(Process):

    def ports_schema(self) -> dict:
        return {}

    def next_update(self, timestep: float, states: dict) -> dict:
        return {}


def serialize_function() -> None:
    pass


class TestSerializer(Serializer):

    def __init__(self, prefix: str = '', suffix: str = '') -> None:
        super().__init__()
        self.prefix = prefix
        self.suffix = suffix

    def can_serialize(self, data: Any) -> bool:
        return isinstance(data, str) and data.startswith('!!')

    def serialize_to_string(self, data: str) -> str:
        return f'{self.prefix}{data}{self.suffix}'

    def deserialize_from_string(self, data: str) -> str:
        if self.suffix:
            return data[len(self.prefix):-len(self.suffix)]
        return data[len(self.prefix):]


def test_serialized_in_serializer_string() -> None:
    serializer = TestSerializer(prefix='![', suffix=']')
    serialized = serializer.serialize('hi there!')
    assert serializer.deserialize(serialized) == 'hi there!'


def test_unmatched_closing_bracket_in_serializer_string() -> None:
    serializer = TestSerializer(prefix='', suffix=']')
    serialized = serializer.serialize('hi there!')
    assert serializer.deserialize(serialized) == 'hi there!'


def test_unmatched_opening_bracket_in_serializer_string() -> None:
    serializer = TestSerializer(prefix='[', suffix='')
    serialized = serializer.serialize('hi there!')
    print(serialized)
    assert serializer.deserialize(serialized) == 'hi there!'


def test_open_bracket_deep_in_serializer_string() -> None:
    serializer = TestSerializer(prefix='abc[', suffix='')
    serialized = serializer.serialize('hi there!')
    assert serializer.deserialize(serialized) == 'hi there!'


def test_close_bracket_deep_in_serializer_string() -> None:
    serializer = TestSerializer(prefix='abc]', suffix='')
    serialized = serializer.serialize('hi there!')
    assert serializer.deserialize(serialized) == 'hi there!'


def test_serialized_prefixing_serializer_string() -> None:
    serializer = TestSerializer(prefix='!TestSerializer[test]', suffix='')
    serialized = serializer.serialize('hi there!')
    assert serializer.deserialize(serialized) == 'hi there!'


def test_exclamation_point_prefixing_serializer_string() -> None:
    serializer = TestSerializer(prefix='!', suffix='')
    serialized = serializer.serialize('hi there!')
    assert serializer.deserialize(serialized) == 'hi there!'


def test_exclamation_point_suffixing_serializer_string() -> None:
    serializer = TestSerializer(prefix='', suffix='!')
    serialized = serializer.serialize('hi there!')
    assert serializer.deserialize(serialized) == 'hi there!'


def test_serialization_full() -> None:
    to_serialize = {
        'process': SerializeProcess(),
        1: True,
        'numpy_int': np.array([1, 2, 3]),
        'numpy_float': np.array([1.1, 2.2, 3.3]),
        'numpy_str': np.array(['a', 'b', 'c']),
        'numpy_bool': np.array([True, True, False]),
        'numpy_matrix': np.array([[1, 2], [3, 4]]),
        'list_units': [1 * units.fg, 2 * units.fg],
        'units_list': [1, 2] * units.fg,
        'list': [True, False, 'test', 1, None],
        'quantity': 5 * units.fg,
        'unit': units.fg,
        'nan_unit': math.nan * units.fg,
        'dict': {
            'a': False,
        },
        'function': serialize_function,
    }
    serialized = serialize_value(to_serialize)
    assert re.fullmatch(
        '!FunctionSerializer\\[<function serialize_function at 0x[0-9a-f]+>\\]',
        serialized.pop('function'))
    expected_serialized = {
        'process': (
            "!ProcessSerializer[{'time_step': 1.0, '_name': "
            "'SerializeProcess'}]"
        ),
        '1': True,
        'numpy_int': [1, 2, 3],
        'numpy_float': [1.1, 2.2, 3.3],
        'numpy_str': ['a', 'b', 'c'],
        'numpy_bool': [True, True, False],
        'numpy_matrix': [[1, 2], [3, 4]],
        'list_units': [
            '!units[1 femtogram]',
            '!units[2 femtogram]',
        ],
        'units_list': [
            '!units[1 femtogram]',
            '!units[2 femtogram]',
        ],
        'list': [True, False, 'test', 1, None],
        'quantity': '!units[5 femtogram]',
        'unit': '!units[femtogram]',
        'nan_unit': '!units[nan femtogram]',
        'dict': {
            'a': False,
        },
    }
    assert serialized == expected_serialized

    # Processes cannot be deserialized.
    expected_serialized.pop('process')

    deserialized = deserialize_value(expected_serialized)

    assert math.isnan(deserialized['nan_unit'].magnitude)
    assert deserialized['nan_unit'].units == units.fg
    deserialized.pop('nan_unit')

    expected_deserialized = {
        '1': True,
        'numpy_int': [1, 2, 3],
        'numpy_float': [1.1, 2.2, 3.3],
        'numpy_str': ['a', 'b', 'c'],
        'numpy_bool': [True, True, False],
        'numpy_matrix': [[1, 2], [3, 4]],
        'list_units': [1 * units.fg, 2 * units.fg],
        'units_list': [1 * units.fg, 2 * units.fg],
        'list': [True, False, 'test', 1, None],
        'quantity': 5 * units.fg,
        'unit': 1 * units.fg,
        'dict': {
            'a': False,
        },
    }
    assert deserialized == expected_deserialized



"""
=====
Store
=====

The file system for storing and updating state variables during an experiment.
"""

import copy
import logging as log
from pprint import pformat
import uuid
import warnings

import numpy as np
from pint import Quantity
from typing import Optional

from vivarium import divider_registry, serializer_registry, updater_registry
from vivarium.core.process import Process, Step
from vivarium.library.dict_utils import deep_merge, deep_merge_check, MULTI_UPDATE_KEY
from vivarium.library.topology import without, dict_to_paths
from vivarium.core.types import Processes, Topology, State, Steps, Flow

_EMPTY_UPDATES = None, None, None, None, None, None
DEFAULT_SCHEMA = '_default'


def generate_state(
        processes: Processes,
        topology: Topology,
        initial_state: Optional[State],
        steps: Optional[Steps] = None,
        flow: Optional[Flow] = None,
) -> 'Store':
    """Initialize a simulation's state.

    Args:
        processes: Simulation processes.
        topology: Topology linking process ports to stores.
        initial_state: Initial simulation state. Omitted variables will
            be assigned values based on schema defaults.

    Returns:
        Initialized state.
    """
    store = Store({})
    steps = steps or {}
    store.generate(tuple(), processes, steps, flow, topology, initial_state)
    store.build_topology_views()

    return store


def view_values(
        states: dict
) -> State:
    state_values = {}
    if isinstance(states, Store):
        return states.get_value()
    for key, value in states.items():
        state_values[key] = view_values(value)
    return state_values


def key_for_value(d, looking):
    """Get the key associated with a value in a dictionary.

    Only top-level keys are searched.

    Args:
        d: The dictionary.
        looking: The value to look for.

    Returns:
        The associated key, or None if no key found.
    """
    found = None
    for key, value in d.items():
        if looking == value:
            found = key
            break
    return found


def hierarchy_depth(hierarchy, path=()):
    """
    Create a mapping of every path in the hierarchy to the node living at
    that path in the hierarchy.
    """

    base = {}

    for key, inner in hierarchy.items():
        down = tuple(path + (key,))
        if isinstance(inner, dict):
            base.update(hierarchy_depth(inner, down))
        else:
            base[down] = inner

    return base


def _always_true(_):
    return True


def _identity(y):
    return y


def topology_path(topology, path):
    '''
    get the subtopology at the path inside the given topology.
    '''

    if not path:
        topology, path
    else:
        head = path[0]
        tail = path[1:]
        if head in topology:
            subtopology = topology[head]
            if isinstance(subtopology, tuple):
                return subtopology, tail
            elif isinstance(subtopology, dict):
                if '_path' in subtopology:
                    if tail and tail[0] in subtopology:
                        down = topology_path(subtopology, tail)
                        if down:
                            return subtopology['_path'] + down[0], down[1]
                    else:
                        return subtopology['_path'], tail
                else:
                    return topology_path(subtopology, tail)

def insert_topology(topology, port_path, target_path):
    assert isinstance(port_path, tuple)
    assert len(port_path) > 0
    head = port_path[0]
    tail = port_path[1:]

    if not tail:
        topology[head] = target_path
    else:
        subtopology = topology[head]
        if isinstance(subtopology, tuple):
            relative_path = tuple(['..' for _ in subtopology]) + target_path
            new_topology = insert_topology(
                {'_path': subtopology},
                tail,
                relative_path)
            topology[head] = new_topology

        elif isinstance(subtopology, dict):
            topology[head] = insert_topology(subtopology, tail, target_path)

        else:
            raise Exception(f'invalid topology {topology} at key {head}')

    return topology

def convert_path(path):
    if isinstance(path, list):
        path = tuple(path)
    elif not isinstance(path, tuple):
        path = (path,)
    return path

class Store:
    """Holds a subset of the overall model state

    The total state of the model can be broken down into :term:`stores`,
    each of which is represented by an instance of this `Store` class.
    The store's state is a set of :term:`variables`, each of which is
    defined by a set of :term:`schema key-value pairs`. The valid schema
    keys are listed in :py:attr:`schema_keys`, and they are:

    * **_default** (Type should match the variable value): The default
      value of the variable.
    * **_updater** (:py:class:`str`): The name of the :term:`updater` to
      use. By default this is ``accumulate``.
    * **_divider** (:py:class:`str`): The name of the :term:`divider` to
      use. Note that ``_divider`` is not included in the ``schema_keys``
      set because it can be applied to any node in the hierarchy, not
      just leaves (which represent variables).
    * **_value** (Type should match the variable value): The current
      value of the variable. This is ``None`` by default.
    * **_properties** (:py:class:`dict`): Extra properties of the
      variable that don't have a specific schema key. This is an empty
      dictionary by default.
    * **_emit** (:py:class:`bool`): Whether to emit the variable to the
      :term:`emitter`. This is ``False`` by default.
    """
    schema_keys = {
        '_default',
        '_updater',
        '_value',
        '_properties',
        '_emit',
        '_serializer',
    }

    def __init__(self, config, outer=None, source=None):
        self.outer = outer
        self.inner = {}
        self.subschema = {}
        self.subtopology = {}
        self.properties = {}
        self.default = None
        self.updater = None
        self.value = None
        self.units = None
        self.divider = None
        self.emit = False
        self.sources = {}
        self.leaf = False
        self.serializer = None
        self.topology = {}
        self.topology_view = None
        # self.flow is None when this node has no flow (either because
        # it is not a Step or because it is a Step treated like a
        # Deriver) and a list when this node holds a Step with
        # dependencies. The list is empty if the Step has no
        # dependencies but should not be treated like a Deriver.
        self.flow = None

        self._apply_config(config, source)

    def __getitem__(self, path):
        '''Retrieve a :term:`hierarchy` node by its :term:`path`.

        .. WARNING:

           This function is **experimental** and part of the
           :term:`store API`.

        Args:
            path: Path, relative to ``self``, of the store to retrieve.

        Returns:
            The store at ``path``. Note that the store is returned, not
            the value of that store.
        '''
        path = convert_path(path)
        return self.get_path(path)

    def __setitem__(self, path, value):
        '''Set a :term:`hierarchy` node's value by its :term:`path`.

        .. WARNING:

           This function is **experimental** and part of the
           :term:`store API`.

        Args:
            path: Path, relative to ``self``, of the store to modify.
            value: The value to be stored as the value of the store at
                ``path``.
        '''
        path = convert_path(path)
        self.set_path(path, value)

    def create(self, path, value=None, absolute=False, **kwargs):
        path = convert_path(path)
        if value:
            kwargs['_value'] = value
        if '_default' not in kwargs:
            kwargs['_default'] = kwargs.get('_value')

        if absolute:
            top = self.top()
            store = top._establish_path(path, config=kwargs)
            top._apply_subschema_path(path)
        else:
            store = self._establish_path(path, config=kwargs)
            self._apply_subschema_path(path)

        store.apply_defaults()
        return store

    def connect(self, path, value, absolute=False):
        '''Wire a store's process to another store.

        This function must not be used unless ``self`` holds a
        :term:`process`.

        .. WARNING:

           This function is **experimental** and part of the
           :term:`store API`.

        Args:
            path: Path of the port to connect.
            value: The store (or the path to the store) to connect to
                the port at ``path``.

        Raises:
            AssertionError: If ``self.value`` is not an instance of
                :py:class:`vivarium.core.process.Process`.
            Exception: If ``value`` is a :py:class:`Store` that is in a
                different tree than ``self``.
        '''
        path = convert_path(path)
        assert isinstance(self.value, Process), \
            f'cannot connect non-process {self.value} at {self.path_for()} to {path}'

        if isinstance(value, Store):
            target_store = value
            if self._independent_store(target_store):
                raise Exception(
                    f"the store being inserted at {path} is from a different tree "
                    f"at {target_store.path_for()}: {target_store.get_value()}")
        else:
            store_path = convert_path(value)
            if absolute:
                target_store = self.top().get_path(store_path)
            else:
                target_store = self.outer.get_path(store_path)

        # update the topology
        self._update_topology(path, target_store)

    def set_path(self, path, value):
        '''Set a value at a path in the hierarchy.

        .. WARNING:

           This function is **experimental** and part of the
           :term:`store API`.

        Args:
            path: The :term:`path` relative to ``self`` where the value
                should be set.
            value: The value to set. The store node at ``path`` will
                hold ``value`` when this function returns.
        '''

        # this case only when called directly
        if len(path) == 0:
            if isinstance(value, Store):
                self.value = value.value
            else:
                self.value = value

        elif len(path) == 1:
            final = path[0]
            if isinstance(value, Store):
                raise Exception(
                    f"the store being inserted at {path} is already in a tree "
                    f"at {value.path_for()}: {value.get_value()}")

                # TODO: make it so a Store can be added here if it has no outer
                # if value.outer:
                #     raise Exception(f"the store being inserted at {path} is already in a tree
                #     at {value.path_for()}: {value.get_value()}")
                # else:
                #     # place the store at this point in the tree
                #     self.inner[final] = value
                #     value.outer = self

            elif isinstance(value, Process):
                self.insert({
                    'processes': value.generate_processes({'name': final}),
                    'topology': value.generate_topology({'name': final}),
                    'initial_state': {},
                })

            else:
                down = self.get_path((final,))
                if down:
                    if not down.leaf:
                        Exception(f'trying to set the value {value} of a branch at {down.path_for()}')
                    down.value = value
                else:
                    Exception(f'trying to set the value {value} at a path that does not exist {final} at {self.path_for()}')

        elif len(path) > 1:
            head = path[0]
            tail = path[1:]
            down = self.get_path((head,))
            if down:
                down.set_path(tail, value)
            else:
                Exception(f'trying to set the value {value} at a path that does not exist {path} at {self.path_for()}')

        else:
            raise Exception("this should never happen")

    def path_to(self, to):
        """return a path from self to the given Store"""

        self_path = self.path_for()
        to_path = to.path_for()
        while len(self_path) > 0 and len(to_path) > 0 and self_path[0] == to_path[0]:
            self_path = self_path[1:]
            to_path = to_path[1:]

        path = [
            '..'
            for _ in self_path]
        path.extend(to_path)
        return tuple(path)

    def _update_topology(self, port_path, target_store):
        """Update the topology with a new port-path pair.

        To use this function ``self`` must hold a :term:`process`.

        .. WARNING:

           This function is **experimental** and part of the
           :term:`store API`.

        Args:
            port_path: Port of the new topology entry.
            target_store: The store to wire to ``port_path``.

        Raises:
            AssertionError: If ``self.value`` is not an instance of
                :py:class:`vivarium.core.process.Process`.
        """

        assert isinstance(self.value, Process), \
            f'assigning topology from {port_path} to {target_store.path_for()} ' \
            f'at {self.path_for()} is invalid, not a process'

        topology = copy.deepcopy(self.topology)
        self.topology = insert_topology(
            topology, port_path, self.outer.path_to(target_store))

        self.value.schema = self.value.get_schema()
        self.outer._topology_ports(
            self.value.schema,
            self.topology,
            source=self.path_for())

        # cache the process's view
        self.topology_view = self.schema_topology(
            self.value.schema,
            self.topology)

    def _independent_store(self, store):
        return self.top() != store.top()

    def _check_default(self, new_default):
        """Check a new default value.

        Compare a new default value to the existing default. If they
        conflict, decide which to rely on and log a warning.

        Returns:
            The new default value.
        """
        defaults_conflict = False
        if self.default is not None:
            self_default_comp = self.default
            new_default_comp = new_default
            if isinstance(self_default_comp, np.ndarray):
                self_default_comp = self.default.tolist()
            if isinstance(new_default_comp, np.ndarray):
                new_default_comp = new_default.tolist()
            defaults_conflict != (self_default_comp == new_default_comp)
        if defaults_conflict:
            if (
                not isinstance(new_default, np.ndarray)
                and not isinstance(self.default, np.ndarray)
                and new_default == 0
                and self.default != 0
            ):
                log.debug(
                    '_default schema conflict: %s and %s. selecting %s',
                    str(self.default), str(new_default), str(self.default))
                return self.default
            log.debug(
                '_default schema conflict: %s and %s. selecting %s',
                str(self.default), str(new_default), str(new_default))
        return new_default

    def _check_schema(self, schema_key, new_schema):
        """Check a new schema value.

        Args:
            schema_key: the schema key
            new_schema: The new schema value.

        Returns:
            The new schema value.

        Raises:
            Exception: If the store schema already has a value and the new
                value is different from the existing one.
        """
        current_schema_value = getattr(self, schema_key)
        if current_schema_value is not None and current_schema_value != new_schema:
            raise ValueError(
                f"Incompatible schema assignment at {self.path_for()}. "
                f"Trying to assign the value {new_schema} to key {schema_key}, "
                f"which already has the value {current_schema_value}.")
        return new_schema

    def _check_schema_support_defaults(self, schema_key, new_schema, schema_registry):
        current_schema_value = getattr(self, schema_key)
        if isinstance(new_schema, str):
            new_schema = schema_registry.access(new_schema)
        if isinstance(new_schema, dict) and isinstance(
                new_schema[schema_key], str):
            new_schema[schema_key] = schema_registry.access(
                new_schema[schema_key])
        if (
                current_schema_value
                and current_schema_value != DEFAULT_SCHEMA
                and current_schema_value != new_schema):
            warnings.warn(
                f"Incompatible schema assignment at {self.path_for()}. "
                f"Trying to assign the value {new_schema} to key {schema_key}, "
                f"which already has the value {current_schema_value}.")
        return new_schema

    def _merge_subtopology(self, subtopology):
        """Merge a new subtopology with the store's existing one."""
        self.subtopology = deep_merge(self.subtopology, subtopology)

    def _apply_subschema_config(self, subschema):
        """Merge a new subschema config with the current subschema."""
        self.subschema = deep_merge(
            self.subschema,
            subschema)

    def _apply_config(self, config, source=None):
        """
        Expand the tree by applying additional config.

        Special keys for the config are:

        * _default - Default value for this node.
        * _properties - An arbitrary map of keys to values. This can be used
          for any properties which exist outside of the operation of the
          tree (like mass or energy).
        * _updater - Which updater to use. Default is 'accumulate' which
          adds the new value to the existing value, but 'set' is common
          as well. You can also provide your own function here instead
          of a string key into the updater library.
        * _emit - whether or not to emit the values under this point in
           the tree.
        * _divider - What to do with this node when division happens.
          Default behavior is to leave it alone, but you can also pass
          'split' here, or a function of your choosing. If you need
          other values from the state you need to supply a dictionary
          here containing the updater and the topology for where the
          other state values are coming from. This has two keys:

            * divider - a function that takes the existing value and any
              values supplied from the adjoining topology.
            * topology - a mapping of keys to paths where the value for
              those keys will be found. This will be passed in as the
              second argument to the divider function.

        * _subschema/* - If this node was declared to house an unbounded set
          of related states, the schema for these states is held in this
          nodes subschema and applied whenever new subkeys are added
          here.
        * _subtopology - The subschema is informed by the subtopology to
          map the process perspective to the actual tree structure.
        * _topology - If this node stores a :term:`process`, then the
          process's topology must be provided under this key. This key
          may only be provided if the node stores a :term:`process`.
        * _flow - If this node stores a :term:`step`, then the step's
          dependencies must be specified under this key as a list of
          :term:`paths` relative to the step's parent node in the
          :term:`hierarchy`. This key must not be provided unless the
          node holds a step.
        """

        if config == '**':
            config = {}  # config needs to be a dict

        # remove _output special key. This is used only by schema_topology.
        config = without(config, '_output')

        if '*' in config:
            self._apply_subschema_config(config['*'])
            config = without(config, '*')

        if '_subschema' in config:
            if source:
                self.sources[source] = config['_subschema']
            self._apply_subschema_config(config['_subschema'])
            config = without(config, '_subschema')

        if '_subtopology' in config:
            self._merge_subtopology(config['_subtopology'])
            config = without(config, '_subtopology')

        if '_topology' in config:
            self.topology = config['_topology']
            config = without(config, '_topology')

        if '_flow' in config:
            flow = config.pop('_flow')
            if flow != {}:
                self.flow = flow

        if '_divider' in config:
            new_divider = config['_divider']
            self.divider = self._check_schema_support_defaults(
                'divider', new_divider, divider_registry)
            config = without(config, '_divider')

        # if emit set in branch, set the entire branch to the emit value
        if '_emit' in config and self.inner:
                emit_value = config['_emit']
                self.set_emit_value(emit=emit_value)
                config = without(config, '_emit')

        if self.schema_keys & set(config.keys()):
            if self.inner:
                raise Exception(
                    'trying to assign leaf values to a branch at: {}'.format(
                        self.path_for()))
            self.leaf = True

            if '_units' in config:
                self.units = self._check_schema(
                    'units', config.get('_units'))
                self.serializer = serializer_registry.access('units')

            if '_serializer' in config:
                serializer = config['_serializer']
                if isinstance(serializer, str):
                    serializer = serializer_registry.access(
                        serializer)
                self.serializer = self._check_schema(
                    'serializer', serializer)

            if '_default' in config:
                self.default = self._check_default(config.get('_default'))
                if isinstance(self.default, Quantity):
                    self.units = self.units or self.default.units
                    self.serializer = (self.serializer or
                                       serializer_registry.access('units'))
                elif isinstance(self.default, list) and \
                        len(self.default) > 0 and \
                        isinstance(self.default[0], Quantity):
                    self.units = self.units or self.default[0].units
                    self.serializer = (self.serializer or
                                       serializer_registry.access('units'))
                elif isinstance(self.default, np.ndarray):
                    self.serializer = (self.serializer or
                                       serializer_registry.access('numpy'))

            if '_value' in config:
                self.value = self._check_schema(
                    'value', config.get('_value'))
                if isinstance(self.value, Quantity):
                    self.units = self.value.units

            if '_updater' in config:
                new_updater = config['_updater']
                self.updater = self._check_schema_support_defaults(
                    'updater', new_updater, updater_registry)

            # All leaf nodes must have an updater
            self.updater = self.updater or DEFAULT_SCHEMA

            # All leaf nodes must have a divider, even though a divider
            # on a branch node higher in the tree will take precedence.
            self.divider = self.divider or DEFAULT_SCHEMA

            self.properties = deep_merge(
                self.properties,
                config.get('_properties', {}))

            self.emit = config.get('_emit', self.emit)

            if source:
                self.sources[source] = config

        else:
            if self.leaf and config:
                if self.value:
                    raise Exception(
                        f'trying to assign create inner for leaf node: '
                        f'{self.path_for()} with value {self.value}')
                self.leaf = False

            for key, child in config.items():
                if key not in self.inner:
                    self.inner[key] = Store(child, outer=self, source=source)
                else:
                    self.inner[key]._apply_config(child, source=source)

        if self.topology and not isinstance(self.value, Process):
            raise ValueError(
                f'Attempting to create Store at {self.path_for()} '
                f'with topology {self.topology}, which is not allowed '
                f'because the Store value ({self.value}) is not a '
                'Process.')
        if self.flow and not self.value.is_step():
            raise ValueError(
                f'Attempting to create Store at {self.path_for()} '
                f'with flow {self.flow}, which is not allowed because '
                f'the Store value ({self.value}) is not a Step.')

    def _get_updater(self, update):
        """Get the updater to use for an update applied to this store.

        Args:
            update: The update.

        Returns:
            If available, the updater specified in the update. If no
            such updater is specified, return this store's default
            updater. If necessary, retrieves updater from the registry.
        """
        updater = self.updater
        if isinstance(update, dict) and '_updater' in update:
            updater = update['_updater']

        if updater == DEFAULT_SCHEMA:
            # For all nodes, by default we use an 'accumulate' updater.
            return updater_registry.access('accumulate')
        elif isinstance(updater, str):
            updater = updater_registry.access(updater)
        return updater

    def _get_divider(self):
        if self.divider == DEFAULT_SCHEMA:
            if self.topology:
                # For processes, we use a null divider by default.
                return divider_registry.access('null')
            # For all other nodes, by default we use a 'set' divider.
            return divider_registry.access('set')
        return self.divider

    def get_config(self, sources=False):
        """
        Assemble a dictionary representation of the config for this node.
        A desired property is that the node can be exactly recreated by
        applying the resulting config to an empty node again.
        """

        config = {}

        if self.properties:
            config['_properties'] = self.properties
        if self.subschema:
            config['_subschema'] = self.subschema
        if self.subtopology:
            config['_subtopology'] = self.subtopology
        if self.divider and self.divider != DEFAULT_SCHEMA:
            config['_divider'] = self.divider

        if sources and self.sources:
            config['_sources'] = self.sources

        if self.inner:
            child_config = {
                key: child.get_config(sources)
                for key, child in self.inner.items()}
            config.update(child_config)

        else:
            config.update({
                '_default': self.default,
                '_value': self.value})
            if self.updater:
                config['_updater'] = self.updater
            if self.units:
                config['_units'] = self.units
            if self.emit:
                config['_emit'] = self.emit

        return config

    def top(self):
        """
        Find the top of this tree.
        """

        if self.outer:
            return self.outer.top()
        return self

    def path_for(self):
        """
        Find the path to this node.
        """

        if self.outer:
            key = key_for_value(self.outer.inner, self)
            above = self.outer.path_for()
            return above + (key,)
        return tuple()

    def get_value(self, condition=None, f=None):
        """
        Pull the values out of the tree in a structure symmetrical to the tree.
        """

        if self.inner:
            if condition is None:
                condition = _always_true

            if f is None:
                f = _identity

            return {
                key: f(child.get_value(condition, f))
                for key, child in self.inner.items()
                if condition(child)}
        if self.subschema:
            return {}
        elif self.topology:
            # this is a process, return it with the topology
            return (self.value, self.topology)
        return self.value

    def get_processes(self):
        """
        Get all processes in this store. Does not include steps.
        """

        if self.inner:
            inner_processes = {}
            for key, child in self.inner.items():
                if child.inner:
                    child_processes = child.get_processes()
                    if child_processes:
                        inner_processes[key] = child_processes
                elif (
                        isinstance(child.value, Process)
                        and not child.value.is_step()):
                    inner_processes[key] = child.value
            if inner_processes:
                return inner_processes
        elif isinstance(self.value, Process):
            return self.value
        return None

    def get_steps(self):
        """Get all steps under this store."""
        if self.inner:
            inner_processes = {}
            for key, child in self.inner.items():
                if child.inner:
                    child_processes = child.get_steps()
                    if child_processes:
                        inner_processes[key] = child_processes
                elif (
                        isinstance(child.value, Process)
                        and child.value.is_step()):
                    inner_processes[key] = child.value
            if inner_processes:
                return inner_processes
        elif isinstance(self.value, Process):
            return self.value
        return None

    def get_topology(self):
        """
        Get the topology for all processes in this store.
        """

        if self.inner:
            inner_topology = {}
            for key, child in self.inner.items():
                child_topology = child.get_topology()
                if child_topology:
                    inner_topology[key] = child_topology
            if inner_topology:
                return inner_topology
        elif self.topology:
            return self.topology
        return None

    def get_flow(self):
        """Get the flow for all :term:`steps` under this node.

        For example:

        >>> from vivarium.core.store import Store
        >>> from vivarium.core.process import Step
        >>> class MyStep(Step):
        ...     def ports_schema(self):
        ...         return {
        ...             'port': ['variable'],
        ...         }
        ...     def next_update(self, timestep, states):
        ...         return {}
        >>> schema = {
        ...     'agent1': {
        ...         'store': {
        ...             'variable': {
        ...                 '_default': 0,
        ...             },
        ...         },
        ...         'step1': {
        ...             '_value': MyStep(),
        ...             '_topology': {
        ...                 'port': ('store',),
        ...             },
        ...             '_flow': [],
        ...         },
        ...         'step2': {
        ...             '_value': MyStep(),
        ...             '_topology': {
        ...                 'port': ('store',),
        ...             },
        ...             '_flow': [('step1',)],
        ...         },
        ...     },
        ... }
        >>> store = Store(schema)
        >>> store.get_flow()
        {'agent1': {'step1': [], 'step2': [('step1',)]}}
        """
        if self.inner:
            inner_flow = {}
            for key, child in self.inner.items():
                child_flow = child.get_flow()
                if child_flow is not None:
                    inner_flow[key] = child_flow
            if inner_flow:
                return inner_flow
        elif self.flow is not None:
            return self.flow
        return None

    def get_path(self, path):
        """
        Get the node at the given path relative to this node.
        """

        if path:
            step = path[0]
            if step == '..':
                child = self.outer
            else:
                child = self.inner.get(step)

            if child:
                return child.get_path(path[1:])
            elif isinstance(self.value, Process):
                towards = topology_path(self.topology, path)
                if towards:
                    target = self.outer.get_path(towards[0])
                    return target.get_path(towards[1])
            else:
                raise Exception(
                    f'{path} is not a valid path from '
                    f'{self.path_for()}')
        return self

    def get_paths(self, paths):
        """Get the nodes at each of the specified paths.

        Args:
            paths: Map from keys to paths.

        Returns:
            A dictionary with the same keys as ``paths``. Each key is
            mapped to the Store object at the associated path.
        """
        return {
            key: self.get_path(path)
            for key, path in paths.items()}

    def get_values(self, paths):
        """Get the values at each of the provided paths.

        Args:
            paths: Map from keys to paths.

        Returns:
            A dictionary with the same keys as ``paths``. Each key is
            mapped to the value at the associated path.
        """
        return {
            key: self.get_in(path)
            for key, path in paths.items()}

    def get_in(self, path):
        """Get the value at ``path`` relative to this store."""
        return self.get_path(path).get_value()

    def get_template(self, template):
        """
        Pass in a template dict with None for each value you want to
        retrieve from the tree!
        """

        state = {}
        for key, value in template.items():
            child = self.inner[key]
            if value is None:
                state[key] = child.get_value()
            else:
                state[key] = child.get_template(value)
        return state

    def emit_data(self):
        """Emit the value at this Store.

        Obeys the schema (namely emits only if ``_emit`` is true). Also
        applies serializers and converts units as necessary.

        Returns:
            The value to emit, or None if nothing should be emitted.
        """
        data = {}
        if self.inner:
            for key, child in self.inner.items():
                child_data = child.emit_data()
                if child_data is not None or child_data == 0:
                    data[key] = child_data
            return data
        if self.emit:
            if self.serializer:
                if isinstance(self.value, list) and self.units:
                    return self.serializer.serialize(
                        [v.to(self.units) for v in self.value])
                if self.units:
                    return self.serializer.serialize(
                        self.value.to(self.units))
                if isinstance(self.value, list):
                    return self.value
                return self.serializer.serialize(self.value)
            if self.units:
                return self.value.to(self.units).magnitude
            return self.value
        return None

    def set_emit_values(self, paths=None, emit=False):
        """
        Turn on/off emits for all inner nodes of the list of paths.
        """
        if paths:
            for path in paths:
                self.set_emit_value(emit=emit, path=path)

    def set_emit_value(self, path=None, emit=False):
        """
        Turn on/off emits for all inner nodes of path.
        """
        if path:
            assert isinstance(path, tuple), 'path must be a tuple'
            target = self.get_path(path)
            target.set_emit_value(emit=emit)
        elif self.inner:
            for child in self.inner.values():
                child.set_emit_value(emit=emit)
        else:
            self.emit = emit

    def _delete_path(self, path):
        """
        Delete the subtree at the given path.
        """

        if not path:
            self.inner = {}
            self.value = None
            return self
        target = self.get_path(path[:-1])
        remove = path[-1]
        if remove in target.inner:
            lost = target.inner[remove]
            del target.inner[remove]
            return lost
        return None

    def divide_value(self):
        """
        Apply the divider for each node to the value in that node to
        assemble two parallel divided states of this subtree.
        """
        divider = self._get_divider()
        if divider:
            # divider is either a function or a dict with topology and/or config
            if isinstance(divider, dict):
                divider_dict = divider
                divider = divider['divider']
                if isinstance(divider, str):
                    divider_str = divider
                    divider = divider_registry.access(divider_str)
                    if not divider:
                        raise Exception(
                            f'No divider found with name {divider_str}')
                args = {}
                if 'topology' in divider_dict:
                    topology = divider_dict['topology']
                    args.update({'state': self.topology_state(topology)})
                if 'config' in divider_dict:
                    config = divider_dict['config']
                    args.update({'config': config})

                return divider(self.get_value(), **args)
            return divider(self.get_value())
        if self.inner:
            daughters = [{}, {}]
            for key, child in self.inner.items():
                division = child.divide_value()
                if division:
                    for daughter, divide in zip(daughters, division):
                        daughter[key] = divide
            return daughters
        return None

    def _reduce(self, reducer, initial=None):
        """
        Call the reducer on each node accumulating over the result.
        """

        value = initial

        for path, node in self.depth():
            value = reducer(value, path, node)
        return value

    def set_value(self, value):
        """
        Set the value for the given tree elements directly instead of using
        the updaters from their nodes.
        """

        if self.inner or self.subschema:
            if not isinstance(value, dict):
                raise Exception(f"trying to set branch {self.path_for()} to value {value}")

            for child, inner_value in value.items():
                if child not in self.inner:
                    if self.subschema:
                        self.inner[child] = Store(self.subschema, self)
                    else:
                        pass
                        # TODO: continue to ignore extra keys?

                if child in self.inner:
                    self.inner[child].set_value(inner_value)
        else:
            self.value = value

    def generate_value(self, value):
        """
        generate the structure for this value that don't exist,
        but don't overwrite any existing values.
        """

        if self.inner or self.subschema:
            if not isinstance(value, dict):
                raise Exception(f"trying to set branch {self.path_for()} to value {value}")

            for child, inner_value in value.items():
                if child not in self.inner:
                    if self.subschema:
                        self.inner[child] = Store(self.subschema, self)
                    else:
                        self._establish_path((child,), {})

                if child in self.inner:
                    self.inner[child].generate_value(inner_value)
        else:
            if self.value is None:
                self.value = value

    def apply_defaults(self):
        """
        If value is None, set to default.
        """
        if self.inner:
            for child in self.inner.values():
                child.apply_defaults()
        else:
            if self.value is None:
                self.value = self.default

    def add(self, added):
        key = added['key']
        inner_keys = self.inner.keys()
        if key in inner_keys:
            raise Exception(
                f"cannot add '{key}' to the hierarchy, "
                f"already present at path {self.path_for()} ")
        elif key == '_unique_id':
            path = (str(uuid.uuid1()),)
        else:
            path = (key,)
        added_state = added['state']

        # get path
        target = self._establish_path(path, {})
        self._apply_subschema_path(path)
        target.apply_defaults()
        target.set_value(added_state)

    def move(self, move, process_store):
        '''
        .. WARNING:

           While this function is used internally by :py:class:`Store`,
           using it as a public method is **experimental** and part of
           the :term:`store API`.
        '''
        process_updates = []
        step_updates = []
        flow_updates = []
        topology_updates = []
        deletions = []

        # get the source node
        source_key = move['source']
        source_path = (source_key,)
        source_node = self.get_path(source_path)

        # move source node to target path
        target_port = move['target']
        target_topology = process_store.topology[target_port]
        target_node = process_store.outer.get_path(target_topology)
        target = target_node.add_node(source_path, source_node)
        target_path = target.path_for() + source_path

        # find the paths to all the processes
        source_process_paths = source_node.depth(
            filter_function=lambda x: isinstance(x.value, Process))

        # find the process and topology updates
        for path, process in source_process_paths:
            process_path = target_path + path
            process_updates.append((
                process_path, process.value))
            topology_updates.append((
                process_path, process.topology))
            if process.value.is_step():
                step_updates.append((
                    process_path, process.value))
                # Note that process.flow may be None, indicating no
                # flow.
                flow_updates.append((
                    process_path, process.flow))

        self._delete_path(source_path)

        here = self.path_for()
        source_absolute = tuple(here + source_path)
        deletions.append(source_absolute)

        return (
            process_updates, step_updates, flow_updates,
            topology_updates, deletions)

    def insert(self, insertion):
        '''
        .. WARNING:

           While this function is used internally by :py:class:`Store`,
           using it as a public method is **experimental** and part of
           the :term:`store API`.
        '''
        process_updates = []
        step_updates = []
        flow_updates = []
        topology_updates = []

        key = insertion.get('key')
        path = (key,) if key else tuple()
        here = self.path_for()

        self.generate(
            path,
            insertion['processes'],
            insertion.get('steps', {}),
            insertion.get('flow'),
            insertion['topology'],
            insertion['initial_state'])

        root = here + path
        process_paths = dict_to_paths(root, insertion['processes'])
        process_updates.extend(process_paths)

        step_paths = dict_to_paths(root, insertion.get('steps', {}))
        step_updates.extend(step_paths)

        topology_paths = [
            (root + (key,), topology)
            for key, topology in insertion['topology'].items()]
        topology_updates.extend(topology_paths)

        flow_paths = [
            (root + (key,), flow)
            for key, flow in insertion.get('flow', {}).items()]
        flow_updates.extend(flow_paths)

        self._apply_subschema_path(path)
        self.get_path(path).apply_defaults()

        return process_updates, step_updates, flow_updates, topology_updates

    def divide(self, divide):
        '''
        .. WARNING:

           While this function is used internally by :py:class:`Store`,
           using it as a public method is **experimental** and part of
           the :term:`store API`.
        '''
        process_and_step_updates = []
        process_updates = []
        step_updates = []
        flow_updates = []
        topology_updates = []
        deletions = []

        # use dividers to find initial states for daughters
        mother = divide['mother']
        mother_path = (mother,)
        daughters = divide['daughters']
        daughter_states = self.inner[mother].divide_value()

        here = self.path_for()

        for daughter, daughter_state in \
                zip(daughters, daughter_states):
            # use initial state as default, merge in divided values
            merged_initial_state = deep_merge(
                daughter_state, daughter.get('initial_state', {}))

            daughter_key = daughter['key']
            daughter_path = (daughter_key,)

            # get the daughter processes
            if 'processes' in daughter or 'steps' in daughter:
                processes = daughter['processes']
                deep_merge_check(processes, daughter.get('steps', {}))
            else:
                # if no processes provided, copy the mother's processes
                mother_processes = self.get_path(mother_path).get_processes()
                processes = copy.deepcopy(mother_processes)
                processes = processes or {}

            # get the daughter topology
            if 'topology' in daughter:
                topology = daughter['topology']
            else:
                # if no topology provided, copy the mother's topology
                mother_topology = self.get_path(mother_path).get_topology()
                topology = copy.deepcopy(mother_topology)
                topology = topology or {}

            # get the daughter flow
            if 'flow' in daughter:
                flow = daughter['flow']
            else:
                # if no flow provided, copy the mother's flow
                mother_flow = self.get_path(mother_path).get_flow()
                flow = copy.deepcopy(mother_flow)
                flow = flow or {}

            self.generate(
                daughter_path,
                processes,
                {},
                flow,
                topology,
                merged_initial_state)

            root = here + daughter_path
            process_paths = dict_to_paths(root, processes)
            process_and_step_updates.extend(process_paths)

            flow_paths = dict_to_paths(root, flow)
            flow_updates.extend(flow_paths)

            topology_paths = [
                (root + (key,), ports)
                for key, ports in topology.items()]
            topology_updates.extend(topology_paths)

            self._apply_subschema_path(daughter_path)
            target = self.get_path(daughter_path)
            target.apply_defaults()
            target.set_value(merged_initial_state)


        self._delete_path(mother_path)
        deletions.append(tuple(here + mother_path))

        for path, process in process_and_step_updates:
            if process.is_step():
                step_updates.append((path, process))
            else:
                process_updates.append((path, process))

        return (
            process_updates, step_updates, flow_updates,
            topology_updates, deletions)

    def delete(self, key, here=None):
        '''
        .. WARNING:

           While this function is used internally by :py:class:`Store`,
           using it as a public method is **experimental** and part of
           the :term:`store API`.
        '''
        if here is None:
            here = self.path_for()
        deletions = []
        path = (key,)
        self._delete_path(path)
        deletions.append(tuple(here + path))

        return deletions

    def apply_update(self, update, state=None):
        """
        Given an arbitrary update, map all the values in that update
        to their positions in the tree where they apply, and update
        these values using each node's `_updater`.

        Arguments:
            update: The update being applied.
            state: The state at the start of the time step.

        There are five topology update methods, which use the following
        special update keys:

        * `_add` - Adds states into the subtree, given a list of dicts
            containing:

            * path - Path to the added state key.
            * state - The value of the added state.

        * `_move` - Moves a node from a source to a target location in the
          tree. This uses an update to an :term:`outer` port, which
          contains both the source and target node locations. Can move
          multiple nodes according to a list of dicts containing:

            * source - the source path from an outer process port
            * target - the location where the node will be placed.

        * `_generate` - The value has four keys, which are essentially
          the arguments to the `generate()` function:

            * path - Path into the tree to generate this subtree.
            * processes - Tree of processes to generate.
            * topology - Connections of all the process's `ports_schema()`.
            * initial_state - Initial state for this new subtree.

        * `_divide` - Performs cell division by constructing two new
          daughter cells and removing the mother. Takes a dict with two keys:

            * mother - The id of the mother (for removal)
            * daughters - List of two new daughter generate directives, of the
                same form as the `_generate` value above.

        * `_delete` - The value here is a list of paths (tuples) to delete from
          the tree.


        Additional special update keys for different update operations:

        * `_updater` - Override the default updater with any updater you want.

        * `_reduce` - This allows a reduction over the entire subtree from some
          point downward. Its three keys are:

            * from - What point to start the reduction.
            * initial - The initial value of the reduction.
            * reducer - A function of three arguments, which is called
              on every node from the `from` point in the tree down:

                * value - The current accumulated value of the reduction.
                * path - The path to this point in the tree
                * node - The actual node being visited.

              This function returns the next `value` for the reduction.
              The result of the reduction will be assigned to this point
              in the tree.
        """
        view_expire = False

        if isinstance(update, dict) and MULTI_UPDATE_KEY in update:
            # apply multiple updates to same node
            multi_update = update[MULTI_UPDATE_KEY]
            assert isinstance(multi_update, list)
            for update_value in multi_update:
                self.apply_update(update_value, state)
            return _EMPTY_UPDATES

        if self.inner or self.subschema:
            # Branch update: this node has an inner
            process_updates = []
            step_updates = []
            flow_updates = []
            topology_updates = []
            deletions = []

            update = dict(update)  # avoid mutating the caller's dict

            add_entries = update.pop('_add', None)
            if add_entries is not None:
                # add a list of sub-states
                for added in add_entries:
                    self.add(added)
                view_expire = True

            move_entries = update.pop('_move', None)
            if move_entries is not None:
                # move nodes from source to target path
                for move in move_entries:
                    (
                        move_processes, move_step, move_flow,
                        move_topology, move_deletions
                    ) = self.move(move, state)
                    process_updates.extend(move_processes)
                    step_updates.extend(move_step)
                    flow_updates.extend(move_flow)
                    topology_updates.extend(move_topology)
                    deletions.extend(move_deletions)
                    view_expire = True

            generate_entries = update.pop('_generate', None)
            if generate_entries is not None:
                # generate a list of new processes
                for generate in generate_entries:
                    (
                        insert_processes, insert_steps, insert_flows,
                        insert_topology
                    ) = self.insert(generate)
                    process_updates.extend(insert_processes)
                    step_updates.extend(insert_steps)
                    flow_updates.extend(insert_flows)
                    topology_updates.extend(insert_topology)
                    view_expire = True

            divide = update.pop('_divide', None)
            if divide is not None:
                (
                    divide_processes, divide_steps, divide_flow,
                    divide_topology, divide_deletions
                ) = self.divide(divide)
                process_updates.extend(divide_processes)
                step_updates.extend(divide_steps)
                flow_updates.extend(divide_flow)
                topology_updates.extend(divide_topology)
                deletions.extend(divide_deletions)
                view_expire = True

            delete_keys = update.pop('_delete', None)

            for key, value in update.items():
                if key in self.inner:
                    inner = self.inner[key]
                    (
                        inner_topology, inner_processes, inner_steps,
                        inner_flows, inner_deletions, inner_view_expire
                    ) = inner.apply_update(value, state)

                    if inner_topology:
                        topology_updates.extend(inner_topology)
                    if inner_processes:
                        process_updates.extend(inner_processes)
                    if inner_steps:
                        step_updates.extend(inner_steps)
                    if inner_flows:
                        flow_updates.extend(inner_flows)
                    if inner_deletions:
                        deletions.extend(inner_deletions)
                    if inner_view_expire:
                        view_expire = inner_view_expire

            if delete_keys is not None:
                # delete a list of paths
                here = self.path_for()
                for key in delete_keys:
                    delete_deletions = self.delete(key, here)
                    deletions.extend(delete_deletions)
                view_expire = True

            return (
                topology_updates, process_updates, step_updates,
                flow_updates, deletions, view_expire)

        # Leaf update: this node has no inner

        updater = self._get_updater(update)

        if isinstance(update, dict) and '_reduce' in update:
            reduction = update['_reduce']
            top = self.get_path(reduction.get('from'))
            update = top._reduce(
                reduction['reducer'],
                initial=reduction['initial'])

        if isinstance(update, dict) and \
                self.schema_keys & set(update.keys()):
            if '_updater' in update:
                update = update.get('_value', self.default)

        try:
            self.value = updater(self.value, update)
        except:
            # provide error messages
            if updater is None:
                raise Exception(
                    f"updater is absent at path {self.path_for()} "
                    f"with value {self.value} for update {pformat(update)}")
            else:
                raise Exception(
                    f"failed update at path {self.path_for()} "
                    f"with value {self.value} for update {pformat(update)}")

        return _EMPTY_UPDATES

    def inner_value(self, key):
        """
        Get the value of an inner state
        """

        if key in self.inner:
            return self.inner[key].get_value()
        return None

    def topology_state(self, topology):
        """
        Fill in the structure of the given topology with the values at all
        the paths the topology points at. Essentially, anywhere in the topology
        that has a tuple path will be filled in with the value at that path.

        This is the inverse function of the standalone `inverse_topology`.
        """

        state = {}

        for key, path in topology.items():
            if key == '*':
                if isinstance(path, dict):
                    node, path = self.outer_path(path)
                    for child, child_node in node.inner.items():
                        state[child] = child_node.topology_state(path)
                else:
                    node = self.get_path(path)
                    for child, child_node in node.inner.items():
                        state[child] = child_node.get_value()
            elif isinstance(path, dict):
                node, path = self.outer_path(path)
                state[key] = node.topology_state(path)
            else:
                state[key] = self.get_path(path).get_value()
        return state

    def schema_topology(self, schema, topology):
        """
        Fill in the structure of the given schema with the connected stores
        according to the given topology.
        """

        state = {}

        if self.leaf or schema == '**':
            state = self
        elif not schema.get('_output'):
            for key, subschema in schema.items():
                path = topology.get(key)
                if key == '*':
                    if isinstance(path, dict):
                        node, path = self.outer_path(path)
                        for child, child_node in node.inner.items():
                            state[child] = child_node.schema_topology(
                                subschema, path)
                    else:
                        node = self.get_path(path)
                        for child, child_node in node.inner.items():
                            state[child] = child_node.schema_topology(
                                subschema, {})
                elif key == '_divider':
                    pass
                elif isinstance(path, dict):
                    node, path = self.outer_path(path)
                    state[key] = node.schema_topology(subschema, path)
                else:
                    if path is None:
                        path = (key,)
                    node = self.get_path(path)
                    if node:
                        state[key] = node.schema_topology(subschema, {})
                    else:
                        # node is None, it was likely deleted
                        print('{} is None'.format(path))

        return state

    def state_for(self, path, keys):
        """
        Get the value of a state at a given path
        """

        state = self.get_path(path)
        if state is None:
            return {}
        if keys and keys[0] == '*':
            return state.get_value()
        return {
            key: state.inner_value(key)
            for key in keys}

    def depth(self, path=(), filter_function=None):
        """
        Create a mapping of every path in the tree to the node living at
        that path in the tree. An optional `filter` argument is a function
        that can declares the instances that will be returned, for example:
        * filter=lambda x: isinstance(x.value, Process)
        """
        base = []
        if filter_function is None or filter_function(self):
            base += [(path, self)]

        for key, child in self.inner.items():
            down = tuple(path + (key,))
            base += child.depth(down, filter_function)
        return base

    def add_node(self, path, node):
        """ Add a node instance at the provided path """
        target = self._establish_path(path[:-1], {})
        if target.get_value() and path[-1] in target.get_value():
            # this path already exists, update it
            self.apply_update({path[-1]: node.get_value()})
        else:
            node.outer = target
            target.inner.update({path[-1]: node})
        return target

    def outer_path(self, path, source=None):
        """
        Address a topology with the `_path` keyword if present,
        establishing a path to this node and using it as the
        starting point for future path operations.
        """

        node = self
        if '_path' in path:
            node = self._establish_path(
                path['_path'],
                {},
                source=source)
            path = without(path, '_path')

        return node, path

    def _apply_subschema_path(self, path):
        """Apply ``self.subschema`` at ``path``."""
        if path:
            inner = self.inner[path[0]]
            if self.subschema:
                subtopology = self.subtopology or {}
                inner._topology_ports(
                    self.subschema,
                    subtopology,
                    source=self.path_for() + ('*',))
            inner._apply_subschema_path(path[1:])

    def _apply_subschema(self, subschema=None, subtopology=None):
        """
        Apply a subschema to all inner nodes (either provided or from this
        node's personal subschema) as governed by the given/personal
        subtopology.
        """

        if subschema is None:
            subschema = self.subschema
        if subtopology is None:
            subtopology = self.subtopology or {}

        inner = list(self.inner.values())

        for child in inner:
            child._topology_ports(
                subschema,
                subtopology,
                source=self.path_for() + ('*',))

    def _apply_subschemas(self):
        """
        Apply all subschemas from all nodes at this point or lower in the tree.
        """

        if self.subschema:
            self._apply_subschema()
        for child in self.inner.values():
            child._apply_subschemas()

    def _update_subschema(self, path, subschema):
        """
        Merge a new subschema into an existing subschema at the given path.
        """

        target = self.get_path(path)
        if target.subschema is None:
            target.subschema = subschema
        else:
            target.subschema = deep_merge(
                target.subschema,
                subschema)
        return target

    def _establish_path(self, path, config, source=None):
        """
        Create a node at the given path if it does not exist, then
        apply a config to it.

        Paths can include '..' to go up a level (which raises an exception
        if that level does not exist).
        """

        if len(path) > 0:
            path_step = path[0]
            remaining = path[1:]

            if path_step == '..':
                if not self.outer:
                    raise Exception(
                        'outer does not exist for path: {}'.format(path))

                return self.outer._establish_path(
                    remaining,
                    config,
                    source=source)
            elif isinstance(self.value, Process):
                towards = topology_path(self.topology, path)
                if towards:
                    target = self.outer._establish_path(towards[0], config, source=source)
                    return target._establish_path(towards[1], config, source=source)
            else:
                if path_step not in self.inner:
                    self.inner[path_step] = Store(
                        {}, outer=self, source=source)

                return self.inner[path_step]._establish_path(
                    remaining,
                    config,
                    source=source)
        else:
            self._apply_config(config, source=source)
            return self

    def _topology_ports(self, schema, topology, source=None):
        """
        Distribute a schema into the tree by mapping its ports
        according to the given topology.
        """

        source = source or self.path_for()

        if set(schema.keys()) & self.schema_keys:
            self.get_path(topology)._apply_config(schema)
        else:
            mismatch_topology = (
                set(topology.keys()) - set(schema.keys()))
            mismatch_schema = (
                set(schema.keys()) - set(topology.keys()))
            if mismatch_topology:
                raise Exception(
                    'topology for the process {} \n at path {} uses '
                    'undeclared ports: {}'.format(
                        source, self.path_for(), str(mismatch_topology)))
            if mismatch_schema:
                log.info(
                    'process %s has ports that are not included in '
                    'the topology: %s', str(source), str(mismatch_schema))

            for port, subschema in schema.items():
                path = topology.get(port, (port,))

                if port == '*':
                    subschema_config = {
                        '_subschema': subschema}
                    if isinstance(path, dict):
                        node, subpath = self.outer_path(
                            path, source=source)
                        node._merge_subtopology(subpath)
                        node._apply_config(subschema_config)
                    else:
                        node = self._establish_path(
                            path,
                            subschema_config,
                            source=source)
                    node._apply_subschema()
                    node.apply_defaults()

                elif isinstance(path, dict):
                    node, subpath = self.outer_path(
                        path, source=source)

                    node._topology_ports(
                        subschema,
                        subpath,
                        source=source)

                else:
                    self._establish_path(
                        path,
                        subschema,
                        source=source)

    def _generate_paths(
            self, processes: Processes, flow: Flow, topology: Topology):
        """Set up state :term:`hierarchy` with stores.

        Recursively creates the entire state hierarchy rooted at
        ``self``.

        .. WARNING::

           This method is public since it is used by
           :py:func:`generate_state`, but it should not be used from
           outside this module. It is not part of the supported API.

        Args:
            processes: Map from process names to process objects,
                embedded within nested dictionaries representing the
                hierarchy nodes rooted at ``self``. Note that steps
                should be included here too.
            flow: Map from step names to relative paths to step
                dependencies, embedded within nested dictionaries
                representing the hierarchy nodes rooted at ``self``.
            topology: The topology.
        """
        for key, subprocess in processes.items():
            subtopology = topology[key]
            subflow = flow.get(key) if flow else None
            if isinstance(subprocess, Process):
                schema = {
                    '_value': subprocess,
                    '_updater': 'set',
                    '_topology': subtopology,
                    '_serializer': 'process'
                }
                if subflow is not None:
                    schema['_flow'] = subflow
                process_state = Store(schema, outer=self)

                self.inner[key] = process_state

                subprocess.schema = subprocess.get_schema()

                self._topology_ports(
                    subprocess.schema,
                    subtopology,
                    source=self.path_for() + (key,))
            else:
                if key not in self.inner:
                    self.inner[key] = Store({}, outer=self)
                self.inner[key]._generate_paths(
                    subprocess,
                    subflow,
                    subtopology,
                )

    def build_topology_views(self):
        if self.leaf:
            if isinstance(self.value, Process):
                self.topology_view = self.outer.schema_topology(
                    self.value.schema,
                    self.topology)
        else:
            for inner in self.inner.values():
                inner.build_topology_views()

    def generate(self, path, processes, steps, flow, topology, initial_state):
        """
        Generate a subtree of this store at the given path.
        The processes will be mapped into locations in the tree by the
        topology, and once everything is constructed the initial_state
        will be applied.
        """

        target = self._establish_path(path, {})
        target._generate_paths(processes, flow, topology)
        target._generate_paths(steps, flow, topology)
        target._apply_subschemas()
        target.set_value(initial_state)
        target.apply_defaults()



import os
import uuid
import logging as log

from vivarium.core.process import (
    Step
)
from vivarium.core.composer import Composer
from vivarium.core.composition import (
    PROCESS_OUT_DIR,
)
from vivarium.core.engine import (
    Engine, pp
)

# processes
from vivarium.composites.toys import ExchangeA
from vivarium.processes.timeline import TimelineProcess

NAME = 'meta_division'


# functions for generating daughter ids
def daughter_uuid(mother_id):
    return [
        str(uuid.uuid1()),
        str(uuid.uuid1())]


def daughter_phylogeny_id(mother_id):
    return [
        str(mother_id) + '0',
        str(mother_id) + '1']


class MetaDivision(Step):
    name = NAME
    defaults = {
        'initial_state': {},
        'daughter_ids_function': daughter_phylogeny_id}

    def __init__(self, parameters=None):
        super().__init__(parameters)

        self.division = 0

        # must provide a compartment to generate new daughters
        self.agent_id = self.parameters['agent_id']
        self.composer = self.parameters['composer']
        self.daughter_ids_function = self.parameters['daughter_ids_function']

    def initial_state(self, config=None):
        return {}

    def ports_schema(self):
        return {
            'global': {
                'divide': {
                    '_default': False,
                    '_updater': 'set',
                    '_divider': {
                        'divider': 'set_value',
                        'config': {
                            'value': False,
                        }
                    },
                }},
            'agents': {}}

    def next_update(self, timestep, states):
        divide = states['global']['divide']

        if divide:
            daughter_ids = self.daughter_ids_function(self.agent_id)
            daughter_updates = []

            for daughter_id in daughter_ids:
                composer = self.composer.generate({
                    'agent_id': daughter_id})
                daughter_updates.append({
                    'key': daughter_id,
                    'processes': composer['processes'],
                    'topology': composer['topology'],
                    'initial_state': {}})

            log.info(
                'DIVIDE! \n--> MOTHER: {} \n--> DAUGHTERS: {}'.format(
                    self.agent_id, daughter_ids))

            # initial state will be provided by division in the tree
            return {
                'agents': {
                    '_divide': {
                        'mother': self.agent_id,
                        'daughters': daughter_updates}}}
        else:
            return {}


# test
class ToyAgent(Composer):
    defaults = {
        'exchange': {'uptake_rate': 0.1},
        'agents_path': ('..', '..', 'agents')}

    def generate_processes(self, config):
        agent_id = config['agent_id']
        division_config = dict(
            {},
            agent_id=agent_id,
            composer=self)

        return {
            'exchange': ExchangeA(config['exchange']),
            'division': MetaDivision(division_config)}

    def generate_topology(self, config):
        agents_path = config['agents_path']
        return {
            'exchange': {
                'internal': ('internal',),
                'external': ('external',)},
            'division': {
                'global': ('global',),
                'agents': agents_path,
            }
        }


def test_division():
    agent_id = '1'

    # initial state
    initial_state = {
        'agents': {
            agent_id: {
                'internal': {'A': 0},
                'external': {'A': 1},
                'global': {
                    'divide': False
                }
            },
        }
    }

    # timeline triggers division
    time_divide = 5
    time_total = 10
    timeline = [
        (0, {('agents', agent_id, 'global', 'divide'): False}),
        (time_divide, {('agents', agent_id, 'global', 'divide'): True}),
        (time_total, {})]

    # declare the hierarchy
    timeline = TimelineProcess({'timeline': timeline})
    agent = ToyAgent({'agent_id': agent_id})

    composite = agent.generate(path=('agents', agent_id))
    composite.merge(
        processes={'timeline': timeline},
        topology={'timeline': {
            'global': ('global',),
            'agents': ('agents',)}},
        state=initial_state
    )

    # configure experiment
    experiment = Engine(
        composite=composite,
    )

    # run simulation
    experiment.update(time_total)
    output = experiment.emitter.get_data()
    experiment.end()

    # external starts at 1, goes down until death, and then back up
    # internal does the inverse
    assert list(output[time_divide]['agents'].keys()) == [agent_id]
    assert agent_id not in list(output[time_divide + 1]['agents'].keys())
    assert len(output[time_divide]['agents']) == 1
    assert len(output[time_divide + 1]['agents']) == 2

    return output


def run_division():
    out_dir = os.path.join(PROCESS_OUT_DIR, NAME)
    os.makedirs(out_dir, exist_ok=True)
    output = test_division()
    pp(output)


if __name__ == '__main__':
    run_division()
